#!/usr/bin/env python3

import logging
import operator
import os
import unittest
from sys import version_info
import pysam
from .. import loci, parsers, utilities, configuration
from Mikado.transcripts.transcript_methods import splitting
import tempfile
from ..parsers import bed12
from ..utilities.log_utils import create_default_logger
if version_info.minor < 5:
    from sortedcontainers import SortedDict
else:
    from collections import OrderedDict as SortedDict

__author__ = 'Luca Venturini'


class TestSplitMonoexonic(unittest.TestCase):

    logger = utilities.log_utils.create_null_logger("test_mono")
    logger.setLevel(logging.DEBUG)

    def setUp(self):
        
        self.transcript = loci.Transcript()
        self.transcript.chrom = "Chr1"
        self.transcript.start = 1001
        self.transcript.end = 6000
        self.transcript.strand = "+"
        self.transcript.id = "transcript1"
        self.transcript.parent = "gene1"
        self.transcript.source = "Mikado"
        self.transcript.exons = [(1001, 6000)]
        
        self.bed1 = parsers.bed12.BED12()
        self.header = False
        self.bed1.chrom = "transcript1"
        self.bed1.start = 1
        self.bed1.end = 5000
        self.bed1.name = "Bed1"
        self.bed1.score = 0
        self.bed1.strand = "+"
        self.bed1.thick_start = 101
        self.bed1.thick_end = 3001
        self.bed1.block_counts = 1
        self.bed1.block_sizes = [2901]
        self.bed1.block_starts = [101]
        self.bed1.transcriptomic = True
        self.bed1.has_start_codon = True
        self.bed1.has_stop_codon = True
        self.assertFalse(self.bed1.invalid, self.bed1.invalid_reason)

        self.bed2 = parsers.bed12.BED12()
        self.header = False
        self.bed2.chrom = "transcript1"
        self.bed2.start = 1
        self.bed2.end = 5000
        self.bed2.name = "Bed2"
        self.bed2.score = 0
        self.bed2.strand = "+"
        self.bed2.thick_start = 4001
        self.bed2.thick_end = 4900
        self.bed2.block_counts = 1
        self.bed2.block_sizes = [900]
        self.bed2.block_starts = [4001]
        self.bed2.transcriptomic = True
        self.bed2.has_start_codon = True
        self.bed2.has_stop_codon = True
        self.assertFalse(self.bed2.invalid)

        self.transcript.logger = self.logger

        self.transcript.configuration = configuration.configurator.load_and_validate_config(None)
        self.transcript.configuration.pick.chimera_split.blast_check = False
        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"
        self.transcript.configuration.pick.chimera_split.blast_params.min_overlap_duplication = 0.8
        self.transcript.configuration.pick.chimera_split.blast_params.minimal_hsp_overlap = 0.8  # 80%
        self.transcript.configuration.pick.chimera_split.blast_params.hsp_evalue = 0.0001
        self.transcript.configuration.pick.chimera_split.blast_params.evalue = 0.0001
        self.transcript.configuration.pick.orf_loading.minimal_secondary_orf_length = 50
        self.db_folder = tempfile.TemporaryDirectory()
        self.transcript.configuration.db_settings.db = os.path.join(self.db_folder.name, "mikado.db")
        self.transcript.load_orfs([self.bed1, self.bed2])
        self.assertTrue(self.transcript.is_coding)
        self.assertEqual(self.transcript.number_internal_orfs,
                         2, str(self.transcript))
        self.transcript.finalize()

        self.transcript.logger = self.logger

        # Setup faux hits
        hits = []

        hit1 = dict()
        hit1["evalue"] = 10**-6
        hit1["hsps"] = []
        hit1["target"] = "target1"
        hit1["target_length"] = 3000
        hit1["query_start"] = 50
        hit1["query_end"] = 3200
        hit1["evalue"] = 10**(-6)
        hit1["global_identity"] = 100
        hit1["query_aligned_length"] = 3200 - 500
        hit1["target_aligned_length"] = 3000 - 1

        hsp1_1 = dict()
        hsp1_1["hsp_evalue"] = 10**(-6)
        hsp1_1["hsp_bits"] = 1200
        hsp1_1["query_hsp_start"] = 50
        hsp1_1["query_hsp_end"] = 3200
        hsp1_1["target_hsp_start"] = 1
        hsp1_1["target_hsp_end"] = 3000
        hsp1_1["match"] = "|" * (3200 - 500)

        hit1["hsps"] = [hsp1_1]

        hits.append(hit1)

        self.transcript.blast_hits = hits
        self.assertEqual(len(self.transcript.blast_hits), 1)

    def test_simple_split(self):

        """
        Check that the original transcript can be split in two
        :return:
        """

        self.assertIsNotNone(self.transcript.configuration)

        self.transcript.configuration.pick.chimera_split.blast_check = False

        # with self.assertLogs("test_mono", level="DEBUG") as log_split:
        new_transcripts = list(splitting.split_by_cds(self.transcript))

        self.assertEqual(len(new_transcripts), 2, "\n".join(str(_) for _ in new_transcripts))
        self.assertEqual(new_transcripts[0].start, self.transcript.start)
        self.assertEqual(new_transcripts[1].end, self.transcript.end)

        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertFalse(sl.configuration.pick.chimera_split.blast_check)
        self.assertEqual(len(sl.transcripts), 1)
        sl.logger.setLevel("DEBUG")
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    def test_lenient_split(self):

        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"

        cds_boundaries = SortedDict()
        for orf in sorted(self.transcript.loaded_bed12,
                          key=operator.attrgetter("thick_start", "thick_end")):
            cds_boundaries[(orf.thick_start, orf.thick_end)] = [orf]

        self.assertEqual(1,
                         len(splitting.check_split_by_blast(self.transcript, cds_boundaries)))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 1)

    def test_stringent_split(self):

        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "STRINGENT"

        cds_boundaries = SortedDict()
        for orf in sorted(self.transcript.loaded_bed12,
                          key=operator.attrgetter("thick_start", "thick_end")):
            cds_boundaries[(orf.thick_start, orf.thick_end)] = [orf]

        self.assertEqual(1,
                         len(splitting.check_split_by_blast(self.transcript, cds_boundaries)))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 1)

    def test_permissive_split(self):

        self.transcript.logger = self.logger

        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "PERMISSIVE"

        cds_boundaries = SortedDict()
        for orf in sorted(self.transcript.loaded_bed12,
                          key=operator.attrgetter("thick_start", "thick_end")):
            cds_boundaries[(orf.thick_start, orf.thick_end)] = [orf]

        self.assertEqual(2, len(splitting.check_split_by_blast(self.transcript, cds_boundaries)))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    @staticmethod
    def get_second_hit():

        hit1 = dict()
        hit1["evalue"] = 10**-6
        hit1["hsps"] = []
        hit1["target"] = "target2"
        hit1["target_length"] = 2000
        hit1["query_start"] = 3700
        hit1["query_end"] = 4850
        hit1["evalue"] = 10**(-6)
        hit1["global_identity"] = 100
        hit1["query_aligned_length"] = 4850 - 3700
        hit1["target_aligned_length"] = 1149

        hsp1_1 = dict()
        hsp1_1["hsp_evalue"] = 10**(-6)
        hsp1_1["hsp_bits"] = 1200
        hsp1_1["query_hsp_start"] = 3700
        hsp1_1["query_hsp_end"] = 4850
        hsp1_1["target_hsp_start"] = 1
        hsp1_1["target_hsp_end"] = 2000
        hsp1_1["match"] = "|" * (4850 - 3700)

        hit1["hsps"] = [hsp1_1]
        return hit1

    def test_permissive_split_twohits(self):

        hit2 = self.get_second_hit()
        self.transcript.blast_hits.append(hit2)
        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "PERMISSIVE"

        self.assertEqual(2, len(list(splitting.split_by_cds(self.transcript))))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    def test_stringent_split_twohits(self):

        hit2 = self.get_second_hit()
        self.transcript.blast_hits.append(hit2)
        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "STRINGENT"

        self.assertEqual(2, len(list(splitting.split_by_cds(self.transcript))))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    def test_no_splitting_by_source(self):
        self.transcript.source = "foo"
        for sources in [[], [self.transcript.source], ["bar"], ["bar", [self.transcript.source]]]:
            with self.subTest(sources=sources):
                self.transcript.configuration.pick.chimera_split.skip = sources
                if self.transcript.source in sources:
                    final = 1
                else:
                    final = 2
                self.assertEqual(final, len(list(splitting.split_by_cds(self.transcript))))
                sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
                sl.load_all_transcript_data()
                self.assertEqual(len(sl.transcripts), final)

    def test_one_orf(self):
        self.transcript.strip_cds()
        self.transcript.load_orfs([self.bed1])
        self.assertEqual(1, len(list(splitting.split_by_cds(self.transcript))))

    def test_no_hsps(self):
        self.transcript.blast_hits = []
        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"

        logger = utilities.log_utils.create_default_logger("test_no_hsps")
        logger.setLevel("DEBUG")
        self.transcript.logger = logger
        self.assertEqual(2, len(list(splitting.split_by_cds(self.transcript))))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    def test_lenient_split_twohits(self):

        hit2 = self.get_second_hit()
        self.transcript.blast_hits.append(hit2)
        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"

        self.assertEqual(2, len(list(splitting.split_by_cds(self.transcript))))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    @staticmethod
    def get_spanning_hit():

        hit1 = dict()
        hit1["evalue"] = 10**-6
        hit1["hsps"] = []
        hit1["target"] = "target1"
        hit1["target_length"] = 5000
        hit1["query_start"] = 51
        hit1["query_end"] = 5000
        hit1["evalue"] = 10**(-6)
        hit1["global_identity"] = 100
        hit1["query_aligned_length"] = 4950
        hit1["target_aligned_length"] = 4950

        hsp1_1 = dict()
        hsp1_1["hsp_evalue"] = 10**(-6)
        hsp1_1["hsp_bits"] = 1200
        hsp1_1["query_hsp_start"] = 51
        hsp1_1["query_hsp_end"] = 5000
        hsp1_1["target_hsp_start"] = 1
        hsp1_1["target_hsp_end"] = 4950
        hsp1_1["match"] = "|" * 4950

        hit1["hsps"] = [hsp1_1]
        return hit1

    def test_spanning_hit_lenient(self):

        self.transcript.blast_hits = [self.get_spanning_hit()]
        self.transcript.configuration.pick.chimera_split.blast_check = True

        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"

        self.assertEqual(1, len(list(splitting.split_by_cds(self.transcript))))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 1)

    def test_spanning_hit_nocheck(self):
        self.transcript.blast_hits = [self.get_spanning_hit()]
        self.transcript.configuration.pick.chimera_split.blast_check = False
        cds_boundaries = SortedDict()
        for orf in sorted(self.transcript.loaded_bed12,
                          key=operator.attrgetter("thick_start", "thick_end")):
            cds_boundaries[(orf.thick_start, orf.thick_end)] = [orf]

        self.assertEqual(len(cds_boundaries), 2)
        self.assertEqual(self.transcript.number_internal_orfs, 2)
        self.assertEqual(2, len(list(splitting.split_by_cds(self.transcript))))
        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    def test_deleted_hits(self):

        delattr(self.transcript, "blast_hits")
        self.transcript.configuration.pick.chimera_split.blast_check = True
        self.transcript.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"
        self.transcript.logger = self.logger
        with self.assertLogs(logger=self.logger, level="WARNING") as log_split:
            self.assertEqual(2, len(list(splitting.split_by_cds(self.transcript))))

        self.assertIn("WARNING:test_mono:BLAST hits store lost for transcript1! Creating a mock one to avoid a crash",
                      log_split.output)

    def test_duplication(self):

        pass


class TestWithPhase(unittest.TestCase):

    logger = utilities.log_utils.create_default_logger("test_mono")
    logger.setLevel(logging.INFO)

    def setUp(self):

        self.transcript = loci.Transcript()
        self.transcript.chrom = "Chr1"
        self.transcript.start = 1001
        self.transcript.end = 6000
        self.transcript.strand = "."
        self.transcript.id = "transcript1"
        self.transcript.parent = "gene1"
        self.transcript.source = "Mikado"
        self.transcript.exons = [(1001, 6000)]
        self.transcript.logger = self.logger
        self.transcript.logger.setLevel("INFO")

        self.transcript.configuration = configuration.configurator.load_and_validate_config("")
        self.transcript.configuration.pick.chimera_split.blast_check = False
        self.assertIsNotNone(self.transcript.configuration)
        self.db_folder = tempfile.TemporaryDirectory()
        self.transcript.configuration.db_settings.db = os.path.join(self.db_folder.name, "mikado.db")

    def testPositive(self):

        self.bed1 = parsers.bed12.BED12()
        self.header = False
        self.bed1.chrom = "transcript1"
        self.bed1.start = 1
        self.bed1.end = 5000
        self.bed1.name = "Bed1"
        self.bed1.score = 0
        self.bed1.strand = "+"
        self.bed1.thick_start = 1
        self.bed1.thick_end = 3002
        self.bed1.phase = 2
        self.bed1.block_counts = 1
        self.bed1.block_sizes = [3002]
        self.bed1.block_starts = [1]
        self.bed1.transcriptomic = True
        self.bed1.has_start_codon = False
        self.bed1.has_stop_codon = True
        self.assertFalse(self.bed1.invalid, self.bed1.invalid_reason)

        self.bed2 = parsers.bed12.BED12()
        self.header = False
        self.bed2.chrom = "transcript1"
        self.bed2.start = 1
        self.bed2.end = 5000
        self.bed2.name = "Bed2"
        self.bed2.score = 0
        self.bed2.strand = "+"
        self.bed2.thick_start = 4001
        self.bed2.thick_end = 4900
        self.bed2.block_counts = 1
        self.bed2.block_sizes = [900]
        self.bed2.block_starts = [4001]
        self.bed2.transcriptomic = True
        self.bed2.has_start_codon = True
        self.bed2.has_stop_codon = True
        self.assertFalse(self.bed2.invalid)

        self.transcript.load_orfs([self.bed1, self.bed2])
        self.assertTrue(self.transcript.is_coding)
        self.assertEqual(self.transcript.number_internal_orfs,
                         2, str(self.transcript))
        self.transcript.finalize()

        # with self.assertLogs("test_mono", level="DEBUG") as log_split:
        new_transcripts = list(splitting.split_by_cds(self.transcript))

        # self.assertIn("DEBUG:test_mono:",
        #               log_split.output)
        self.assertEqual(len(new_transcripts), 2, "\n".join(str(_) for _ in new_transcripts))
        self.assertEqual(new_transcripts[0].start, self.transcript.start)
        self.assertEqual(new_transcripts[1].end, self.transcript.end)

    def testNegative(self):
        self.bed1 = parsers.bed12.BED12()
        self.header = False
        self.bed1.chrom = "transcript1"
        self.bed1.start = 1
        self.bed1.end = 5000
        self.bed1.name = "Bed1"
        self.bed1.score = 0
        self.bed1.strand = "-"
        self.bed1.thick_start = 1
        self.bed1.thick_end = 3000
        self.bed1.phase = 0
        self.bed1.block_counts = 1
        self.bed1.block_sizes = [3001]
        self.bed1.block_starts = [1]
        self.bed1.transcriptomic = True
        self.bed1.has_start_codon = True
        self.bed1.has_stop_codon = True
        self.assertFalse(self.bed1.invalid, self.bed1.invalid_reason)

        logger = create_default_logger("testNegative", "DEBUG")
        self.bed2 = parsers.bed12.BED12(logger=logger)
        self.header = False
        self.bed2.chrom = "transcript1"
        self.bed2.start = 1
        self.bed2.end = 5000
        self.bed2.name = "Bed2"
        self.bed2.score = 0
        self.bed2.strand = "-"
        self.bed2.thick_start = 4001
        self.bed2.thick_end = 5000
        self.bed2.phase = 1
        self.bed2.block_counts = 1
        self.bed2.block_sizes = [1000]
        self.bed2.block_starts = [4001]
        self.bed2.transcriptomic = True
        self.bed2.has_start_codon = False
        self.bed2.has_stop_codon = True
        self.assertFalse(self.bed2.invalid, (self.bed2.phase, self.bed2.invalid_reason))

        self.transcript.load_orfs([self.bed1, self.bed2])
        self.assertTrue(self.transcript.is_coding)
        self.assertEqual(self.transcript.number_internal_orfs,
                         2,
                         str(self.transcript))
        self.transcript.finalize()

        self.assertEqual(self.transcript.number_internal_orfs,
                         2,
                         str(self.transcript))

        # with self.assertLogs("test_mono", level="DEBUG") as log_split:
        new_transcripts = list(splitting.split_by_cds(self.transcript))

        self.assertEqual(len(new_transcripts), 2, "\n".join(str(_) for _ in new_transcripts))
        self.assertEqual(new_transcripts[0].start, self.transcript.start)
        self.assertEqual(new_transcripts[1].end, self.transcript.end)

        self.assertEqual(self.transcript.combined_cds_start, 6000)
        self.assertEqual(self.transcript.combined_cds_end, 1001)
        self.assertEqual(self.transcript.selected_cds_end, 1001)
        self.assertEqual(self.transcript.selected_cds_start, 4000)
        self.assertEqual(self.transcript.strand, "-")

        sl = loci.Superlocus(self.transcript, configuration=self.transcript.configuration)
        self.assertEqual(len(sl.transcripts), 1)
        sl.load_all_transcript_data()
        self.assertEqual(len(sl.transcripts), 2)

    def test_negative_orf_gtg(self):

        fasta = """>sex_morph_FW.stringtie_sex_morph_FW_str.232.2
CACAGTCTCGTGCGGCTATTTTCGTCCGCCGCCTGTCCCTCTAAGAAGAGTTTAAGCTCC
TGGGAGCCGGCGGTAGCCCTAGTAACGTATCGTGATCCGCCGGCGACGGCCGCGAACGCG
GCGCGCTTCACCACGGAGCCCCAGGCACGACAGACGCACGCCGACCAGGGGATGAACCCC
GGCCGACGCACGCCCGCCGCACGCAGGGACGCGGATGGCGCGGCCGCGCCCGACGACCGC
CGTGGACGACGGGCGAACGCGTTCGGGGATACCGGGCCGAGCCGACGGGAACGCGAACAC
GGACGGCCGAAACCGCCCGCGCCGCGCCCACCGCCGACCCGGGTTTACCCGCCTAGTTAG
CAGGACAGAGTCTCGTTCGTTATCGGAATTAACCAGACAGATCGCTCCACCAACTAAGAA
CGGCCATGCACCACCACCCACCGAATCAAGAAAGAGCTCTCAATCTGTCAATCTTTCCGG
TGTCCGGGCCTGGTGAGGTTTCCCGTGTTGAGTCAAATTAAGCCGCAGGCTCCACTCCTG
GTGGTGCCCTTCCGTCAATTCCTTTAAGTTTCAACTTTGCAATCATACTTCCCCCGGAAC
CGAAAAGCTTCGGTTTCCCGGAAGCTGCCCGCCGGGTCGTTAATGAAACGCCGGCGGATC
GCTAGCTGGCATCGTTTACAGTTAGAACTAGGGCGGTATCTGATCGCCTTCGAACCTCTA
ACTTTCGTTCTTGATCATACGAGAACGTACTTGGCAAATGCTTTCGCGTCAGTTCGTCTC
GAGACGATCCAAGAATTTCACCTCTAACGTCTCGGTACGAATGCCCCCGCCCGTCTCTGT
TGATCATTACCCCGGAGGGCGATTTCGCGCGCCCGCGAAGGGCGGAGATGCGCGGGACCA
AGGTCTTGTTCCATTATTCCATGCGACCAGTATTCAGGGCCTTTTGACGAGACGGCCGTG
AAGCCGCCCCGCCAGATTTGAGCCTGCTTTGAGCACTCTAATTTGTTCAAAGTAAACGTG
TCGGCCCGCCGACGGCACTCGGTGAAGAGCACCGCGCAGCAAGATTGGAGTAGGCGGCCG
CCGTCGTCGAACCCCGACGGCCGCGCGACGCGTGGCCGCGCGGCGCGCCGGAAGCACGAG
ACACGTGTCCGCC"""

        prodigal = "\t".join(["sex_morph_FW.stringtie_sex_morph_FW_str.232.2",
                              "Prodigal_v2.6.3",
                              "CDS", "934","1137", "5.7", "-", "0",
                              "ID=12199_3;partial=00;start_type=GTG;rbs_motif=None;rbs_spacer=None;gc_cont=0.657;conf=78.73;score=5.69;cscore=7.02;sscore=-1.33;rscore=0.33;uscore=-0.91;tscore=-0.75;"
                              ])
        with tempfile.NamedTemporaryFile(mode="wt", suffix=".fa") as temp_fa, \
                tempfile.NamedTemporaryFile(mode="wt", suffix=".gff3") as temp_gff:
            print("##gff-version\t3", file=temp_gff)
            print(prodigal, file=temp_gff)
            print(fasta, file=temp_fa)
            temp_gff.flush()
            temp_fa.flush()
            handle = open(temp_gff.name, mode="rt")
            fasta_index = pysam.FastaFile(temp_fa.name)
            bed12_parser = bed12.Bed12Parser(handle,
                                             fasta_index=fasta_index,
                                             is_gff=True,
                                             transcriptomic=True,
                                             max_regression=1)
            record = next(bed12_parser)
            self.assertIsInstance(record, bed12.BED12)
            self.assertFalse(record.has_start_codon)
            self.assertFalse(record.invalid)
            self.assertEqual(record.phase, 1, record)
            os.remove(temp_fa.name + ".fai")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3


import tempfile

import pkg_resources

from Mikado.exceptions import InvalidParsingFormat
from ..parsers import blast_utils
import unittest
import os
import gzip
import subprocess
from ..serializers.blast_serializer import xml_utils as seri_blast_utils
from ..serializers.blast_serializer.tabular_utils import matrices
from ..serializers.blast_serializer import parse_btop
import numpy as np
import time
import itertools
from pytest import mark
from collections import namedtuple
from shutil import which


class BtopTester(unittest.TestCase):

    def test_btop_equals(self):
        matrix=matrices["blosum62"]
        for qmult, tmult in itertools.product([1, 3], [1, 3]):
            for qsize, ssize in itertools.product(range(10 * qmult, 20 * qmult, 10),
                                                  range(10 * tmult, 20 * tmult, 10)):
                for qpos, spos in itertools.product(range(0, qsize, 2), range(0, ssize, 2)):
                    mqlength = (qsize - qpos) // qmult
                    mslength = (ssize - spos) // tmult
                    for mlength in range(1, int(min(mqlength, mslength))):
                        with self.subTest(qsize=qsize, ssize=ssize, qpos=qpos, spos=spos, mlength=mlength):
                            qar, sar = np.zeros([3, qsize], dtype=int), np.zeros([3, ssize], dtype=int)
                            sm = str(mlength)
                            qar, sar, tot, match = parse_btop(sm, qpos, spos, qar, sar, matrix, qmult=qmult, tmult=tmult)
                            qfound = np.where(qar > 0)
                            sfound = np.where(sar > 0)
                            self.assertEqual(tot, mlength * min(qmult, tmult),
                                             (tot, mlength, qpos, spos, qmult, tmult))
                            self.assertTrue(((qfound[1][qfound[0] == 0] == qfound[1][qfound[0] == 1]) &
                                 (qfound[1][qfound[0] == 0] == qfound[1][qfound[0] == 2])).all()
                            )
                            self.assertTrue(((sfound[1][sfound[0] == 0] == sfound[1][sfound[0] == 1]) &
                                 (sfound[1][sfound[0] == 0] == sfound[1][sfound[0] == 2])).all()
                            )
                            self.assertEqual(qfound[1][qfound[0] == 0].shape[0], mlength * qmult)
                            self.assertEqual(sfound[1][sfound[0] == 0].shape[0], mlength * tmult,
                                             (tot, tmult, sar, sfound[1][sfound[0] == 0]))
                            # self.assertEqual(qfound[0].shape[0], 3)
                            self.assertEqual(np.where(qar[0] > 0)[0].min(), qpos,
                                             (qfound[0], qpos, mlength))
                            self.assertEqual(np.where(qar[0] > 0)[0].max(), qpos + mlength * qmult - 1,
                                             (qar[0], qpos + tot * qmult, tot))
                            self.assertEqual(np.where(sar[0] > 0)[0].min(), spos)
                            self.assertEqual(np.where(sar[0] > 0)[0].max(), spos + mlength * tmult - 1)

    def test_only_match(self):
        for qmult, tmult in itertools.product([1, 3], [1, 3]):
            for qsize, ssize in itertools.product(range(10 * qmult, 20 * qmult, 10),
                                                  range(10 * tmult, 20 * tmult, 10)):
                for qpos, spos in itertools.product(range(0, qsize, 2), range(0, ssize, 2)):
                    mqlength = (qsize - qpos) // qmult
                    mslength = (ssize - spos) // tmult
                    for mlength in range(1, int(min(mqlength, mslength))):
                        for score in (-1, 0, 1):
                            with self.subTest():
                                match = "AT" * mlength
                                qar, sar = np.zeros([3, qsize], dtype=int), np.zeros([3, ssize], dtype=int)
                                qar, sar, tot, match = parse_btop(match, qpos, spos, qar, sar, {"AT": score},
                                                           qmult=qmult, tmult=tmult)
                                qfound = np.where(qar > 0)
                                sfound = np.where(sar > 0)
                                self.assertEqual(tot, mlength * min(qmult, tmult),
                                                 (tot, mlength, qpos, spos, qmult, tmult))
                                self.assertTrue(qfound[1][qfound[0] == 0].shape[0] == mlength * qmult)
                                self.assertTrue(sfound[1][sfound[0] == 0].shape[0] == mlength * tmult)
                                self.assertTrue(qfound[1][qfound[0] == 1].shape[0] == 0)
                                self.assertTrue(sfound[1][sfound[0] == 1].shape[0] == 0)
                                if score <= 0:
                                    self.assertTrue(qfound[1][qfound[0] == 2].shape[0] == 0)
                                    self.assertTrue(sfound[1][sfound[0] == 2].shape[0] == 0)
                                else:  # Positives found!
                                    self.assertTrue((qfound[1][qfound[0] == 0] == qfound[1][qfound[0] == 2]).all(),
                                                    (score, qfound))
                                    self.assertTrue((sfound[1][sfound[0] == 0] == sfound[1][sfound[0] == 2]).all())
                                # self.assertEqual(qfound[0].shape[0], 3)
                                self.assertEqual(np.where(qar[0] > 0)[0].min(), qpos, (qfound[0], qpos, mlength))
                                self.assertEqual(np.where(qar[0] > 0)[0].max(), qpos + mlength * qmult - 1)
                                self.assertEqual(np.where(sar[0] > 0)[0].min(), spos)
                                self.assertEqual(np.where(sar[0] > 0)[0].max(), spos + mlength * tmult - 1)

    @mark.slow
    def test_gap(self):

        for qmult, tmult in itertools.product([1, 3], [1, 3]):
            for qsize, ssize in itertools.product(range(10 * qmult, 20 * qmult, 5),
                                                  range(10 * tmult, 20 * tmult, 5)):
                for qpos, spos in itertools.product(range(0, qsize, 5), range(0, ssize, 5)):
                    mqlength = (qsize - qpos) // qmult
                    mslength = (ssize - spos) // tmult
                    for mlength in range(3, int(min(mqlength, mslength))):
                        for gap_pos in range(mlength - 1):
                            for gap in ("A-", "-A"):
                                sm = ""
                                if gap_pos - 1 > -1:
                                    sm += str(gap_pos)
                                sm += gap
                                if mlength - gap_pos - 1:
                                    sm += str(mlength - gap_pos)
                                qar, sar = np.zeros([3, qsize], dtype=int), np.zeros([3, ssize], dtype=int)
                                qar, sar, tot, match = parse_btop(sm, qpos, spos, qar, sar, dict(), qmult=qmult, tmult=tmult)
                                qfound = np.where(qar > 0)
                                sfound = np.where(sar > 0)
                                self.assertEqual(tot, mlength * min(qmult, tmult) + min(qmult, tmult),
                                                 (sm, tot, mlength, gap_pos, qpos, spos, qmult, tmult))
                                if gap[0] == "-":  # gap in query
                                    qadd, sadd = 0, tmult
                                    self.assertIn(spos + gap_pos * tmult, np.where(sar[1] == 0)[0],
                                                  (sm, mlength, spos, gap_pos, tmult, sar))
                                else:
                                    qadd, sadd = qmult, 0
                                    self.assertIn(qpos + gap_pos * qmult, np.where(qar[1] == 0)[0])

                                self.assertEqual(np.where(qar[0] > 0)[0].shape[0], mlength * qmult + qadd)
                                self.assertEqual(np.where(sar[0] > 0)[0].shape[0], mlength * tmult + sadd,
                                                 (sar, sm))
                                self.assertEqual(np.where(sar[1] > 0)[0].shape[0], mlength * tmult)
                                self.assertEqual(np.where(qar[1] > 0)[0].shape[0], mlength * qmult)


class XMLLineTester(unittest.TestCase):

    class HSP(object):
        seq_t = namedtuple("seq", ["seq"])

        def __init__(self, query=None, target=None, mid=None,
                     query_start=0, query_end=0, ident_num=0, pos_num=0,
                     frame=0):
            self.aln_annotation = dict()
            self.query_start, self.query_end = query_start, query_end
            self.ident_num, self.pos_num = ident_num, pos_num
            self.query_frame = frame
            self.set_seqs(query, target, mid)

        def set_seqs(self, query, target, mid):
            self.query = self.seq_t(query)
            self.hit = self.seq_t(target)
            self.aln_annotation["similarity"] = mid

    def test_aln_string_basic(self):
        _qlenght, _slength = 20, 20
        for qpos, spos in itertools.product(range(1, _qlenght), range(1, _qlenght)):
            for qmult, tmult in itertools.product((1, 3), (1, 3)):
                qlength, slength = _qlenght * qmult, _slength * tmult
                max_length = min((qlength - qpos) // qmult,
                                 (slength - spos) // tmult)
                for l in range(1, max_length + 1):
                    qseq, sseq, mid = "A" * l, "A" * l, "A" * l
                    hs = self.HSP(query=qseq, target=sseq, mid=mid,
                                  query_start=qpos, query_end=qpos + l * qmult,
                                  ident_num=l, pos_num=l,
                                  frame=0)
                    match, ident, pos = seri_blast_utils.prepare_aln_strings(hs, qmultiplier=qmult)
                    self.assertEqual(match, "|" * l)
                    self.assertTrue((ident == pos).all())


class BlastBasics(unittest.TestCase):

    def test_sniff_correct(self):

        valid_xml = os.path.join(
            os.path.dirname(__file__),
            "mikado.blast.xml"
        )

        valid, header, exc = blast_utils.BlastOpener(valid_xml).sniff()
        self.assertTrue(valid, (valid, exc))
        self.assertIsNone(exc, exc)

        with open(valid_xml, mode="rt") as new_handle:
            valid, header, exc = blast_utils.BlastOpener(new_handle).sniff()
            self.assertTrue(valid, (valid, exc))
            self.assertIsNone(exc, exc)

    def test_sniff_invalid(self):

        with tempfile.NamedTemporaryFile(delete=True) as invalid_xml:
            invalid_xml.write(b"failing\n")
            invalid_xml.flush()
            with self.assertRaises((ValueError, InvalidParsingFormat)):
                _ = blast_utils.BlastOpener(invalid_xml.name).sniff()

    def test_sniff_inexistent(self):

        with tempfile.NamedTemporaryFile(delete=True) as inexistent:
            pass

        with self.assertRaises(OSError):
            _ = blast_utils.BlastOpener(inexistent.name).sniff()

    @mark.slow
    def test_sniff_gzip(self):

        with tempfile.NamedTemporaryFile(suffix=".xml.gz") as handle:
            new = gzip.open(handle.name, mode="wb")

            with pkg_resources.resource_stream("Mikado.tests", "mikado.blast.xml") as valid_xml:
                for line in valid_xml:
                    new.write(line)
            new.flush()

            valid, header, exc = blast_utils.BlastOpener(new.name).sniff()
            self.assertTrue(valid, (valid, exc))
            self.assertIsNone(exc, exc)

            with gzip.open(new.name, mode="rt") as new_handle:
                valid, header, exc = blast_utils.BlastOpener(new_handle).sniff()
                self.assertTrue(valid, (valid, exc))
                self.assertIsNone(exc, exc)

    def test_fail_closed(self):

        valid_xml = os.path.join(
            os.path.dirname(__file__),
            "mikado.blast.xml"
        )

        opener = blast_utils.BlastOpener(valid_xml)

        opener.close()
        with self.assertRaises(ValueError):
            opener.sniff()

    @unittest.skipUnless(which("blast_formatter") is not None and which("makeblastdb") is not None,
                         "This test can be performed only if NCBI+ is installed")
    @mark.slow
    def test_asn(self):

        folder = tempfile.TemporaryDirectory()

        # valid_asn = tempfile.NamedTemporaryFile(suffix=".asn", mode="wb")
        valid_asn = os.path.join(folder.name, "mikado.blast.asn")
        comp_asn = os.path.join(folder.name, "mikado.blast.asn.gz")
        with pkg_resources.resource_stream("Mikado.tests", "mikado.blast.asn.gz") as compressed_asn, \
                open(valid_asn, "wb") as valid_out, open(comp_asn, "wb") as comp_out:
            stream = compressed_asn.read()
            comp_out.write(stream)
            valid_out.write(gzip.decompress(stream))

        uni_out = os.path.join(folder.name, "uniprot_sprot_plants.fasta")
        with pkg_resources.resource_stream("Mikado.tests", "uniprot_sprot_plants.fasta.gz") as uni, \
                open(uni_out, "wb") as uni_handle:
            uni_handle.write(gzip.decompress(uni.read()))

        master = os.getcwd()
        os.chdir(folder.name)
        subprocess.call(
            "makeblastdb -in {uni_out} -dbtype=prot".format(uni_out=uni_out),
            shell=True, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)

        valid, header, exc = blast_utils.BlastOpener(valid_asn).sniff()
        self.assertTrue(valid, (valid, exc))
        self.assertIsNone(exc, exc)

        valid, header, exc = blast_utils.BlastOpener("mikado.blast.asn.gz").sniff()
        self.assertTrue(valid, (valid, exc))
        self.assertIsNone(exc, exc)
        with blast_utils.BlastOpener(comp_asn) as gzasn:
            self.assertFalse(gzasn.closed)
            self.assertTrue(gzasn.open)
            _ = next(gzasn)

        # This should fix the issue of blast_formatter crashing
        time.sleep(1)
        os.chdir(master)


class TestMerging(unittest.TestCase):

    """Small class to test basic cases of the merging algorithm."""

    def test_merging_1(self):

        l = [(-10, -5), (-6, 8), (5, 10), (20, 40)]
        tot_length = 51
        corr_merged = [(-10, 10), (20, 40)]
        merged, tot_length = seri_blast_utils.merge(l, query_length=tot_length, offset=1)
        self.assertTrue((merged == corr_merged))
        self.assertEqual(tot_length, 10 - -10 +1 + 40 - 20 + 1)

    def test_merging_2(self):

        l = [(100, 200)]
        for offset in [0, 1, 2]:
            with self.subTest(offset=offset):
                tot_length = l[0][1] - l[0][0] + offset
                if offset == 2:
                    with self.assertRaises(ValueError):
                        _ = seri_blast_utils.merge(l, offset=offset)
                else:
                    merged, length = seri_blast_utils.merge(l, offset=offset)
                    self.assertEqual(length, tot_length)
                    self.assertTrue((merged == l), (merged, l))

    def test_various_merging(self):

        invalid = [
            [('a', 0)],
            [('a', 'b')],
            [(10, 20), ('a', 'b')],
            [(10, 20, 30), (40, 50, 60)]
        ]

        for inv in invalid:
            with self.subTest(inv=inv):
                with self.assertRaises(TypeError, msg=inv):
                    seri_blast_utils.merge(inv)

        valid = {
            0: [[('10', '20')], [(10, 20)]],
            1: [[(10, 30)], [(10, 30)]],
            2: [[(-10.0, 5.5)], [(-10, 5)]],
            3: [[(-4, -10)], [(-10, -4)]],
            4: [[(-5, -10), (-2.2, -7.3)], [(-10, -2)]]
        }

        for val in valid:
            inp, out = valid[val]
            with self.subTest(val=val, msg=valid[val]):
                _ = seri_blast_utils.merge(inp)
                self.assertTrue((out == _[0]), (out, _[0]))

    def test_included(self):

        cases = {
            tuple([(10, 60), (40, 100), (200, 400)]): [(10, 100), (200, 400)],
            tuple([(54, 1194), (110, 790), (950, 1052)]): [(54, 1194)],
            tuple([(54, 1194), (110, 790), (950, 1052), (1200, 1400)]): [(54, 1194), (1200, 1400)]
        }
        for val, out in cases.items():
            with self.subTest(val=val, msg=cases[val]):
                _ = seri_blast_utils.merge(list(val))
                self.assertTrue((out == _[0]), (out, _[0]))

    def test_unordered(self):
        cases = {
            tuple([(10, 60), (40, 100), (200, 400)]): [(10, 100), (200, 400)],
            tuple([(54, 1194), (110, 790), (950, 1052)]): [(54, 1194)],
            tuple([(54, 1194), (110, 790), (950, 1052), (1200, 1400)]): [(54, 1194), (1200, 1400)]
        }

        from random import shuffle
        for num in range(30):
            for val, out in cases.items():
                cval = list(val[:])
                shuffle(cval)
                with self.subTest(val=val, cval=cval, msg=cases[val]):
                    _ = seri_blast_utils.merge(list(cval))
                    self.assertTrue((out == _[0]), (out, _[0]))


if __name__ == '__main__':
    unittest.main()


# coding: utf-8

"""
This module defines the classes needed for the "compare" script, namely:

- result_storer:    a glorified struct to
                    hold the comparison results of a prediction vs. the reference
- reference_gene:   data structure which holds the transcripts of a gene.
                    No check is performed - transcripts are grouped according
                    to their parent information.
- assigner:         Main workhorse of the sublibrary. This class
                    assigns each transcript to its best match(es) in the reference annotation.
- accountant:       This class calculates the final summary statistics for the comparison.
"""

from . import resultstorer
from . import compare
from .contrast import compare as c_compare



#!/usr/bin/env python3

"""
This subprogram is the second step in the pipeline. Its purpose is to
create the database that is needed by pick for fast access to the data
necessary during the analysis.
"""

import argparse
import functools
import glob
import os
import shutil
import sys
import logging
import logging.handlers
from copy import copy
from typing import Union

import tempfile

import io

from Mikado import version

from ..utilities import path_join, comma_split, create_shm_handle
from ..utilities.dbutils import DBBASE
from ..utilities.log_utils import create_default_logger, create_null_logger
from ._utils import check_log_settings_and_create_logger
from ..utilities import dbutils
from ..exceptions import InvalidConfiguration
from ..exceptions import InvalidSerialization
import random
from ..utilities import blast_keys
from ..configuration import MikadoConfiguration, DaijinConfiguration
import pysam
import sqlalchemy
from ..configuration import configurator


__author__ = 'Luca Venturini'


def setup_shm_db(configuration: MikadoConfiguration, logger=create_null_logger()) -> (MikadoConfiguration,
                                                                                      Union[None, io.FileIO]):
    """
    This method will copy the SQLite input DB into memory.
    """

    if configuration.serialise.shm is True and configuration.db_settings.dbtype == "sqlite":
        handle = create_shm_handle()
        if handle is None:
            logger.info("Mikado was asked to copy the database into /dev/shm, but it is either " 
                        "not available or not writable for this user.")
            configuration.serialise.shm = False
            return configuration, None
        else:
            logger.info(
                f"Using {handle.name} as temporary database before copying into {configuration.db_settings.db}.")
            configuration.db_settings.db = handle.name
            return configuration, handle


def xml_launcher(xml_candidate=None, configuration=None, logger=None):

    """
    Thin rapper around blast_utils.XmlSerializer. Its purpose is
    to create a standard serializer object (use it with partial
     from functools)

    :param xml_candidate: An XML or ASN BLAST file name

    :param configuration: the configuration dictionary, if available
    :type configuration: (None | MikadoConfiguration|DaijinConfiguration)

    :param logger: the logger instance.
    :type logger: (None | logging.Logger)

    :return:
    """

    from ..serializers import blast_serializer
    xml_serializer = blast_serializer.BlastSerializer(
        xml_candidate,
        configuration=configuration,
        logger=logger)
    xml_serializer()


def load_junctions(mikado_configuration: Union[MikadoConfiguration, DaijinConfiguration], logger):
    """
    Function that performs the loading of the junctions.

    :param mikado_configuration: Mikado configuration object

    :param logger: the logging instance.
    :type logger: (None | logging.Logger)

    :return:
    """

    assert isinstance(mikado_configuration, (MikadoConfiguration, DaijinConfiguration))

    if not mikado_configuration.serialise.files.junctions:
        logger.info("Skipping junction loading as no junctions have been provided.")
        return

    if not mikado_configuration.reference.genome:
        exc = InvalidConfiguration(
            "Missing the genome FAI file for serialising the junctions. \
I cannot proceed with this step!")
        logger.exception(exc)
        raise exc

    logger.info("Starting to load junctions: %s",
                mikado_configuration.serialise.files.junctions)
    for junction_file in iter(
            j_file for j_file in mikado_configuration.serialise.files.junctions
            if j_file != ''):
        logger.debug("Loading junctions: %s", junction_file)
        from ..serializers import junction
        serializer = junction.JunctionSerializer(
            junction_file,
            configuration=mikado_configuration,
            logger=logger)
        serializer()
    logger.info("Loaded junctions")


def load_blast(mikado_configuration: Union[MikadoConfiguration, DaijinConfiguration], logger):

    """
    Function to load the BLAST data into the chosen database.

    :param mikado_configuration: Mikado configuration object

    :param logger: the logging instance.
    :type logger: (None | logging.Logger)

    """
    if mikado_configuration.serialise.files.xml:
        logger.info("Starting to load BLAST data")
        filenames = []

        part_launcher = functools.partial(
            xml_launcher,
            **{"configuration": mikado_configuration, "logger": logger})

        for xml in mikado_configuration.serialise.files.xml:
            if os.path.isdir(xml):
                filenames.extend(
                    [os.path.join(xml, _xml) for _xml in
                     os.listdir(xml) if _xml.endswith(
                        (".xml", ".asn", ".tsv", ".txt", ".xml.gz", ".asn.gz", ".tsv.gz", ".txt.gz")) is True])
            else:
                filenames.extend(glob.glob(xml))

        if len(filenames) > 0:
            part_launcher(filenames)
            logger.info("Finished to load BLAST data")
        else:
            logger.warning("No valid BLAST file specified, skipping this phase")


def load_orfs(mikado_configuration: Union[DaijinConfiguration, MikadoConfiguration], logger):

    """
    Function to load the ORFs into the DB.
    :param mikado_configuration: configuration object
    :param logger:
    :return:
    """

    if len(mikado_configuration.serialise.files.orfs) > 0:
        from ..serializers import orf
        logger.info("Starting to load ORF data")
        for orf_file in mikado_configuration.serialise.files.orfs:
            logger.debug("Starting to load ORFs from %s", orf_file)
            try:
                serializer = orf.OrfSerializer(orf_file,
                                               configuration=mikado_configuration,
                                               logger=logger)
                serializer()
            except InvalidSerialization as exc:
                logger.critical("Mikado serialise failed due to problems with the input data. Please check the logs.")
                raise exc
        logger.info("Finished loading ORF data")
    else:
        logger.info("No ORF data provided, skipping")


def load_external(mikado_configuration: Union[MikadoConfiguration, DaijinConfiguration], logger):

    """Function to load external data from."""

    if mikado_configuration.serialise.files.external_scores in (None, ""):
        logger.debug("No external scores to load, returning")
        return
    else:
        logger.info("Starting to load external data")
        from ..serializers import external
        with external.ExternalSerializer(
                mikado_configuration.serialise.files.external_scores,
                configuration=mikado_configuration,
                logger=logger) as serializer:
            serializer()
        logger.info("Finished loading external data")


def _set_serialise_files_options(conf: Union[MikadoConfiguration, DaijinConfiguration],
                                 args, logger=create_null_logger()) -> Union[MikadoConfiguration, DaijinConfiguration]:
    conf.serialise.files.orfs = args.orfs.split(",") if args.orfs else conf.serialise.files.orfs
    conf.serialise.files.xml = args.xml.split(",") if args.xml else conf.serialise.files.xml
    conf.serialise.files.junctions = args.junctions.split(",") if args.junctions else conf.serialise.files.junctions
    conf.serialise.files.transcripts = args.transcripts if args.transcripts else conf.serialise.files.transcripts
    conf.serialise.files.blast_targets = args.blast_targets if args.blast_targets else \
        conf.serialise.files.blast_targets
    conf.reference.genome_fai = args.genome_fai if args.genome_fai else \
        conf.reference.genome_fai
    conf.reference.genome = args.genome if args.genome else conf.reference.genome

    conf.serialise.files.output_dir = args.output_dir if args.output_dir is not None else \
        conf.serialise.files.output_dir

    try:
        os.makedirs(conf.serialise.files.output_dir, exist_ok=True)
    except (OSError, PermissionError, FileExistsError) as exc:
        logger.error("Failed to create the output directory!")
        logger.exception(exc)
        raise exc
        
    if args.db is not None:
        conf.db_settings.db = args.db
        conf.db_settings.dbtype = "sqlite"    

    if conf.serialise.files.output_dir != "." and conf.db_settings.dbtype == "sqlite":
        conf.db_settings.db = path_join(conf.serialise.files.output_dir,
                                        os.path.basename(conf.db_settings.db))

    if conf.serialise.files.junctions:
        if conf.reference.genome_fai in (None, ""):
            if conf.reference.genome not in (None, ""):
                _ = pysam.Fastafile(conf.reference.genome)
                conf.reference.genome_fai = conf.reference.genome + ".fai"
            else:
                exc = InvalidConfiguration("Missing FAI file for junction loading!")
                logger.critical(exc)
                raise exc

    return conf


def _set_serialise_run_options(conf: Union[MikadoConfiguration, DaijinConfiguration],
                               args, logger=create_null_logger()) -> Union[MikadoConfiguration, DaijinConfiguration]:
    conf.serialise.max_regression = args.max_regression or conf.serialise.max_regression
    conf.serialise.start_adjustment = args.start_adjustment
    conf.serialise.max_target_seqs = args.max_target_seqs or conf.serialise.max_target_seqs
    conf.serialise.single_thread = args.single_thread or conf.serialise.single_thread
    conf.serialise.shm = args.use_shm if args.use_shm is not None else conf.serialise.shm
    conf.serialise.files.blast_loading_debug = True if args.blast_loading_debug else \
        conf.serialise.files.blast_loading_debug
    conf.serialise.force = args.force if args.force is not None else conf.serialise.force
    # File with the external scores
    conf.serialise.files.external_scores = args.external_scores if args.external_scores else \
        conf.serialise.files.external_scores
    conf.serialise.codon_table = args.codon_table if isinstance(args.codon_table, (int, str)) else \
        conf.serialise.codon_table
    return conf


def _execute_force_removal(configuration: Union[MikadoConfiguration, DaijinConfiguration],
                           logger=create_null_logger()):

    if configuration.db_settings.dbtype == "sqlite":
        if os.path.exists(configuration.db_settings.db):
            logger.warn("Removing old data from %s because force option in place",
                        configuration.db_settings.db)
            os.remove(configuration.db_settings.db)
            assert not os.path.exists(configuration.db_settings.db)
        else:
            pass
        return

    engine = dbutils.connect(configuration)
    meta = sqlalchemy.MetaData(bind=engine)
    meta.reflect(engine)
    for tab in reversed(meta.sorted_tables):
        logger.debug("Dropping %s", tab)
        tab.drop()
        if configuration.db_settings.dbtype == "mysql":
            engine.execute("OPTIMIZE TABLE {}".format(tab.name))
    if configuration.db_settings.dbtype != "mysql":
        engine.execute("VACUUM")
    dbutils.DBBASE.metadata.create_all(engine)
    return


def setup(args):

    """
    Function to set up everything for the serialisation.
    :param args:
    :return:
    """

    mikado_configuration = configurator.load_and_validate_config(args.configuration)
    logger = create_default_logger("serialiser")
    mikado_configuration.multiprocessing_method = args.start_method if args.start_method is not None else \
        mikado_configuration.multiprocessing_method
    mikado_configuration.threads = args.procs if (args.procs is not None and args.procs > 0) else \
        mikado_configuration.threads

    # Retrieve data from the argparse and put it into the configuration
    mikado_configuration = _set_serialise_files_options(args=args, conf=mikado_configuration, logger=logger)
    mikado_configuration = _set_serialise_run_options(mikado_configuration, args, logger=logger)
    mikado_configuration = configurator.load_and_validate_config(mikado_configuration)
    mikado_configuration, logger = check_log_settings_and_create_logger(mikado_configuration,
                                                                        args.log, args.log_level,
                                                                        section="serialise")
    logger.setLevel("INFO")
    logger.info(f"Mikado version: {version.__version__}")
    logger.info("Command line: %s", " ".join(sys.argv))
    if args.random_seed is True:
        mikado_configuration.seed = None
    elif args.seed is not None:
        mikado_configuration.seed = args.seed
    else:
        pass

    mikado_configuration.check()
    random.seed(mikado_configuration.seed)
    logger.info("Random seed: %s", mikado_configuration.seed)
    logger.setLevel(mikado_configuration.log_settings.log_level)

    # Add sqlalchemy logging
    sql_logger = logging.getLogger("sqlalchemy.engine")
    if args.log_level == "DEBUG":
        level = mikado_configuration.log_settings.log_level
    else:
        level = mikado_configuration.log_settings.sql_level

    sql_logger.setLevel(level)
    sql_logger.addHandler(logger.handlers[0])

    logger.info("Using a %s database (location: %s)", mikado_configuration.db_settings.dbtype,
                mikado_configuration.db_settings.db)

    logger.info("Requested %d threads, forcing single thread: %s", mikado_configuration.threads,
                mikado_configuration.serialise.single_thread)

    if mikado_configuration.serialise.force is True:
        _execute_force_removal(mikado_configuration, logger=logger)
        if (args.force is True or mikado_configuration.serialise.force is True) and mikado_configuration.db_settings.dbtype == "sqlite":
            assert not os.path.exists(mikado_configuration.db_settings.db)

    return mikado_configuration, logger, sql_logger


def serialise(args):

    """
    Wrapper around the serializers objects. It uses the configuration
    supplied through the command line to launch the necessary tools.

    :param args: namespace with the necessary information for the serialisation
    :return:
    """

    mikado_configuration, logger, sql_logger = setup(args)
    intended_db = copy(mikado_configuration.db_settings.db)
    pre_existing = os.path.exists(intended_db)
    if mikado_configuration.serialise.shm is True and mikado_configuration.db_settings.dbtype == "sqlite":
        mikado_configuration, temp_db = setup_shm_db(mikado_configuration, logger=logger)
        if pre_existing and mikado_configuration.serialise.force is False:
            shutil.copy(intended_db, temp_db)
    elif mikado_configuration.db_settings.dbtype == "sqlite":
        temp_db = tempfile.NamedTemporaryFile(suffix=".db")
        mikado_configuration.db_settings.db = temp_db.name
    else:
        temp_db = None

    try:
        load_orfs(mikado_configuration, logger)
        load_blast(mikado_configuration, logger)
        load_external(mikado_configuration, logger)
        load_junctions(mikado_configuration, logger)
    except (ValueError,OSError,PermissionError,IndexError,InvalidSerialization) as exc:
        logger.error("Mikado crashed due to an error. Please check the logs for hints on the cause of the error; if "
                     "it is a bug, please report it to https://github.com/EI-CoreBioinformatics/mikado/issues.")
        logger.error(exc)
        assert (temp_db is None and mikado_configuration.db_settings.dbtype != "sqlite") or (
                temp_db is not None and mikado_configuration.db_settings.dbtype == "sqlite")
        if mikado_configuration.db_settings.dbtype == "sqlite":
            temp_db.close()
            if pre_existing and os.path.exists(intended_db):
                os.remove(intended_db)
        else:
            engine = dbutils.connect(mikado_configuration)
            connection = engine.connect()
            for tbl in reversed(DBBASE.metadata.sorted_tables):
                connection.execute(tbl.delete())
        logging.shutdown()
        sys.exit(1)

    logger.info("Finished")

    if temp_db is not None:
        logger.info(f"Copying from the temporary database {mikado_configuration.db_settings.db} "
                    f"back to disk {intended_db}.")
        shutil.copy(mikado_configuration.db_settings.db, intended_db)
        logger.info(f"Copied from the temporary database {mikado_configuration.db_settings.db} "
                    f"back to disk {intended_db}.")
        temp_db.close()

    logging.shutdown()


def serialise_parser():
    """
    Parser function for the serialisation step.
    :return: argparse.Namespace
    """

    parser = argparse.ArgumentParser("Serialisation utility of the Mikado suite.")
    parser.add_argument("--start-method", dest="start_method",
                        choices=["fork", "spawn", "forkserver"],
                        default=None, help="Multiprocessing start method.")
    shm = parser.add_mutually_exclusive_group()
    shm.add_argument("--shm", action="store_true", default=None, dest="use_shm",
                     help="Use /dev/shm (if available) for faster database building.")
    shm.add_argument("--no-shm", action="store_false", default=None, dest="use_shm",
                     help="Force building the database on its final location, even if /dev/shm is available.")
    orfs = parser.add_argument_group()
    orfs.add_argument("--orfs", type=str, default=None,
                      help="ORF BED file(s), separated by commas")
    orfs.add_argument("--transcripts", default=None,
                      help="""Transcript FASTA file(s) used for ORF calling and BLAST queries,
                      separated by commas.
                      If multiple files are given, they must be in the same order of the
                      ORF files.
                      E.g. valid command lines are:

                      --transcript_fasta all_seqs1.fasta --orfs all_orfs.bed
                      --transcript_fasta seq1.fasta,seq2.fasta --orfs orfs1.bed,orf2.bed
                      --transcript_fasta all_seqs.fasta --orfs orfs1.bed,orf2.bed

                      These are invalid instead:

                      # Inverted order
                      --transcript_fasta seq1.fasta,seq2.fasta --orfs orfs2.bed,orf1.bed
                      #Two transcript files, one ORF file
                      --transcript_fasta seq1.fasta,seq2.fasta --orfs all_orfs.bed
                      """)
    orfs.add_argument("-mr", "--max-regression", dest="max_regression",
                      type=float, default=None,
                      help=""""Amount of sequence in the ORF (in %%) to backtrack
                      in order to find a valid START codon, if one is absent. Default: %(default)s""")
    orfs.add_argument("--codon-table", dest="codon_table", default=None,
                      help="""Codon table to use. Default: 0 (ie Standard, NCBI #1, but only ATG is considered \
a valid start codon.""")
    orfs.add_argument("-nsa", "--no-start-adjustment", default=True,
                      action="store_false",
                      dest="start_adjustment",
                      help="Disable the start adjustment algorithm. Useful when using e.g. TransDecoder vs 5+.")

    blast = parser.add_argument_group()
    blast.add_argument("--max-target-seqs", type=int, default=None,
                       help="Maximum number of target sequences.")
    blast.add_argument("-bt", "--blast-targets", "--blast_targets", default=[], type=comma_split,
                       help="Target sequences")
    blast.add_argument("--xml", "--tsv", type=str, dest="xml", help="""BLAST file(s) to parse.
    They can be provided in three ways:
    - a comma-separated list
    - as a base folder
    - using bash-like name expansion (*,?, etc.). In this case, you have to
    enclose the filename pattern in double quotes.

    Multiple folders/file patterns can be given, separated by a comma.
    BLAST files must be either of two formats:
- BLAST XML
- BLAST tabular format, with the following **custom** fields:
    {fields}
    """.format(fields=" ".join(blast_keys)), default=[])
    blast.add_argument("-p", "--procs", type=int,
                       help="""Number of threads to use for
    analysing the BLAST files. This number should not be higher than the total number of XML files.
    """, default=None)
    blast.add_argument("--single-thread", action="store_true",
                       default=None, dest="single_thread",
                       help="""Force serialise to run with a single thread, irrespective of
                       other configuration options.""")

    junctions = parser.add_argument_group()
    junctions.add_argument("--genome_fai", default=None)
    junctions.add_argument("--genome", default=None)
    junctions.add_argument("--junctions", type=str, default=None)

    external_args = parser.add_argument_group()
    external_args.add_argument(
        "--external-scores",
        dest="external_scores",
        help="""Tabular file containing external scores for the transcripts.
        Each column should have a distinct name, and transcripts have to be listed on the first column.""")

    generic = parser.add_argument_group()
    generic.add_argument("-mo", "--max-objects", dest="max_objects",
                         type=int, default=None,  # So it can actually be set through the JSON
                         help="""Maximum number of objects to cache in memory before
committing to the database. Default: 100,000 i.e. approximately 450MB RAM usage for Drosophila.""")
    forcing = parser.add_mutually_exclusive_group()
    forcing.add_argument("--no-force", action="store_false", default=None, dest="force",
                         help="""Flag. If set, do not drop the contents of an existing Mikado DB
before beginning the serialisation.""")
    forcing.add_argument("--force", action="store_true", default=None, dest="force",
                         help="""Flag. If set, an existing databse will be deleted (sqlite)
or dropped (MySQL/PostGreSQL) before beginning the serialisation.""")
    # If None, the default configuration will be used (from the blueprint)
    generic.add_argument("--configuration", "--json-conf", default=None,
                         dest="configuration", type=str,
                         required=False)
    generic.add_argument("-l", "--log", type=str, default=None, nargs='?', help="Optional log file. Default: stderr")
    parser.add_argument("-od", "--output-dir", dest="output_dir",
                        type=str, default=None,
                        help="Output directory. Default: current working directory")
    log_arguments = parser.add_mutually_exclusive_group()
    log_arguments.add_argument("-lv", "--log-level", default=None,
                         choices=["DEBUG", "INFO", "WARN", "ERROR"],
                         help="Log level. Default: derived from the configuration; if absent, INFO")
    log_arguments.add_argument("--verbose", default=None, dest="log_level", action="store_const", const="DEBUG")
    log_arguments.add_argument("--quiet", default=None, dest="log_level", action="store_const", const="WARNING")
    log_arguments.add_argument("--blast-loading-debug", default=None,
                               dest="blast_loading_debug", action="store_true",
                               help="Flag. If set, Mikado will switch on the debug mode for the XML/TSV loading.")
    generic.add_argument("db", type=str, default=None,
                         nargs='?',
                         help="Optional output database. Default: derived from configuration")
    seed_group = parser.add_mutually_exclusive_group()
    seed_group.add_argument("--seed", type=int, default=None, help="Random seed number. Default: 0.")
    seed_group.add_argument("--random-seed", action="store_true", default=False,
                            help="Generate a new random seed number (instead of the default of 0)")
    parser.set_defaults(func=serialise)
    return parser



#!/usr/bin/env python3

import Mikado
import argparse
import sys
from collections import defaultdict
import operator

__doc__ = """Script to convert a cDNA_match GFF3 from exonerate into either a match/match_part or a GTF file."""

def main():

    parser = argparse.ArgumentParser(__doc__)
    parser.add_argument("-of", "--out-format", choices=["gtf", "match"], default="gtf", dest="format")
    parser.add_argument("gff", nargs="?", default="-")
    parser.add_argument("out", nargs="?", default=sys.stdout)
    args = parser.parse_args()

    if args.gff == "-":
        args.gff = sys.stdin
        pass
    else:
        args.gff = open(args.gff)
        pass
    
    if args.out == "-":
        args.out = sys.stdout
    elif not args.out is sys.stdout:
        args.out = open(args.out, "wt")
    
    matches = defaultdict(list)
    positions = dict()
    
    
    for line in args.gff:
        line = Mikado.parsers.GFF.GffLine(line)
        if line.header is True:
            continue
        matches[line.id].append(line)
        if line.id not in positions:
            positions[line.id] = (line.chrom, line.start, line.end)
        else:
            if positions[line.id][0] != line.chrom:
                raise ValueError("Chromosome mismatch!")
            positions[line.id] = (line.chrom,
                                  min(line.start, positions[line.id][1]),
                                  max(line.end, positions[line.id][2]))
        continue

    rev_positions = defaultdict(list)
    for tid, val in positions.items():
        rev_positions[val].append(tid)
    
    for pos in sorted(rev_positions.keys(), key=operator.itemgetter(0, 1, 2)):
        for match in rev_positions[pos]:
            transcript = Mikado.transcripts.Transcript()
            transcript.chrom, transcript.strand, transcript.id = matches[match][0].chrom, matches[match][0].strand, match
            transcript.attributes["Target"] = matches[match][0].attributes["Target"]
            transcript.add_exons(matches[match])
            transcript.finalize()
            if args.format == "gtf":
                transcript.parent = transcript.id
                print(transcript.format("gtf", with_cds=False), file=args.out)
                pass
            else:
                for line in transcript.format("gff", with_cds=False).split("\n"):
                    line = line.split("\t")
                    if line[2] == "exon":
                        line[2] = "match_part"
                    else:
                        line[2] = "match"
                    print(*line, sep="\t", file=args.out)
                    continue
                pass
            continue
        continue

    args.gff.close()
    args.out.close()
    return

main()



# coding: utf-8

"""
This module defines a containers that hold transcripts excluded from further consideration.
It is invoked when all transcripts in a locus have a score of 0 and the "purge"
option has been enabled.
"""

from .abstractlocus import Abstractlocus
from ..transcripts import Transcript


class Excluded(Abstractlocus):
    """This is a container of discarded transcripts. It is used only for completeness purposes -
    i.e. printing out the discarded transcripts to a separate file.
    """

    __name__ = "excluded_transcripts"

    def __init__(self, monosublocus_instance=None, configuration=None, logger=None):
        """
        Constructor method

        :param monosublocus_instance:
        :type monosublocus_instance: Mikado.loci_objects.monosublocus.Monosublocus

        :param configuration: configuration file
        :type configuration: (MikadoConfiguration|DaijinConfiguration)

        :param logger: logger instance
        :type logger: logging.Logger | None
        """

        Abstractlocus.__init__(self, configuration=configuration)
        self.splitted = False
        self.metrics_calculated = False
        self.logger = logger
        if isinstance(monosublocus_instance, Transcript):
            Abstractlocus.__init__(self, transcript_instance=monosublocus_instance)
        elif isinstance(monosublocus_instance, Abstractlocus):
            # Add the transcript to the Locus
            self.add_monosublocus(monosublocus_instance)

    def add_transcript_to_locus(self, transcript, **kwargs):
        """Override of the sublocus method, and reversal to the original
        method in the Abstractlocus class.
        :param transcript: a transcript to add
        :type transcript: Mikado.loci_objects.transcript.Transcript

        :param kwargs: optional arguments are completely ignored by this method.
        """

        # Notice that check_in_locus is always set to False.
        _ = kwargs

        Abstractlocus.add_transcript_to_locus(self, transcript, check_in_locus=False)

    def add_monosublocus(self, monosublocus_instance):
        """Wrapper to extract the transcript from the monosubloci and pass it
        to the constructor.
        :param monosublocus_instance
        :type monosublocus_instance: Mikado.loci_objects.monosublocus.Monosublocus
        """
        assert len(monosublocus_instance.transcripts) == 1
        for tid in monosublocus_instance.transcripts:
            self.add_transcript_to_locus(monosublocus_instance.transcripts[tid])

    def __str__(self):
        """This special method is explicitly *not* implemented;
        this Locus object is not meant for printing, only for computation!"""
        message = """This is a container used for computational purposes only,
        it should not be printed out directly!"""
        raise NotImplementedError(message)

    def filter_and_calculate_scores(self, check_requirements=True):
        """
        Suppress the method from the base class
        """

        raise NotImplementedError("Scores are not calculated by this class!")

    def define_monosubloci(self):
        """
        Suppress the method from the base class
        """

        raise NotImplementedError("Monosubloci are not calculated by this class!!")

    @classmethod
    def is_intersecting(cls):
        """Present to fulfill the contract with Abstractlocus, but it only raises a NotImplementedError"""
        raise NotImplementedError()



from dataclasses import field

from marshmallow import validate
from marshmallow_dataclass import dataclass, List, Optional, Union


@dataclass
class FilesConfiguration:
    junctions: List[str] = field(default_factory=lambda: [], metadata={
                "metadata": {"description": "File of reliable junctions (from e.g. Portcullis), in BED12 format"},
    })
    xml: List[str] = field(default_factory=lambda: [], metadata={
                "metadata": {
                    "description": "BLAST file or folder of files, in XML or tabular format, optionally compressed."},
    })
    blast_loading_debug: bool = field(default=False, metadata={
                "metadata": {
                    "description": "Boolean flag. If True, the loading of BLAST files will happen \
in a single thread and debugging mode will be activated."},
    })
    external_scores: str = field(default="", metadata={
                "metadata": {
                    "description": "File of additional scores related to the input transcripts. \
Please see the documentation."},
    })
    orfs: List[str] = field(default_factory=lambda: [], metadata={
                "metadata": {
                    "description": "File(s) containing the ORF calling for the input transcripts, in GFF3 or BED12 \
*transcriptomic* (ie in cDNA rather than genomic) coordinates."},
    })
    transcripts: str = field(default="mikado_prepared.fasta", metadata={
                "metadata": {"description": "Input transcripts in FASTA format, ie the output of Mikado prepare."},
                "required": True
    })
    log: str = field(default="serialise.log", metadata={"metadata": {"description": "log file."}})
    blast_targets: List[str] = field(default_factory=lambda: [], metadata={
                "metadata": {"description": "FASTA file(s) with the BLAST targets."},
    })
    output_dir: str = field(default=".", metadata={"validate": validate.Length(min=1)})


@dataclass
class SerialiseConfiguration:
    files: Optional[FilesConfiguration] = field(default_factory=FilesConfiguration, metadata={
                "metadata": {"description": "Options related to input files for serialise"},
    })
    substitution_matrix: str = field(default="blosum62", metadata={
                "metadata": {
                    "description": "Substitution matrix used for the BLAST. This value will be derived from the XML \
files, but it must be provided here or on the command line when using BLAST tabular data. \
Default: blosum62, the default for both BLAST and DIAMOND."},
                "required": True,
                "validate": validate.OneOf(choices=["blosum45", "blosum50", "blosum62",
                                                    "blosum80", "blosum90", "pam250", "pam30", "pam70"])
    })
    blast_flavour: str = field(default="blastx", metadata={
                "metadata": {"description": "Type of BLAST used. Either BLASTP or BLASTX. Default: blastx, \
which should be the sane presumption in most instances."},
                "validate": validate.OneOf(choices=["blastx", "blastp"])
    })
    codon_table: Union[int, str] = field(default=0, metadata={
                "metadata": {
                    "description": "codon table to use for verifying/modifying the ORFs. Default: 0, ie the universal \
codon table but enforcing ATG as the only valid start codon."},
                "required": True,
                "validate": validate.OneOf(
                    choices=[0, 1, 2, 3, 4, 5, 6, 9, 10, 11, 12, 13, 14, 15, 16, 21, 22,
                             23, 24, 25, 26, 27, 28, 29, 30, 31,
                             "Standard", "SGC0",
                             "Vertebrate Mitochondrial", "SGC1",
                             "Yeast Mitochondrial", "SGC2",
                             "Mold Mitochondrial",
                             "Protozoan Mitochondrial",
                             "Coelenterate Mitochondrial", "Mycoplasma",
                             "Spiroplasma", "SGC3",
                             "Invertebrate Mitochondrial", "SGC4",
                             "Ciliate Nuclear", "Dasycladacean Nuclear",
                             "Hexamita Nuclear", "SGC5",
                             "Echinoderm Mitochondrial",
                             "Flatworm Mitochondrial",
                             "SGC8", "Euplotid Nuclear", "SGC9",
                             "Bacterial", "Archaeal", "Plant Plastid",
                             "Alternative Yeast Nuclear",
                             "Ascidian Mitochondrial",
                             "Alternative Flatworm Mitochondrial",
                             "Blepharisma Macronuclear",
                             "Chlorophycean Mitochondrial",
                             "Trematode Mitochondrial",
                             "Scenedesmus obliquus Mitochondrial",
                             "Thraustochytrium Mitochondrial",
                             "Pterobranchia Mitochondrial",
                             "Candidate Division SR1", "Gracilibacteria",
                             "Pachysolen tannophilus Nuclear",
                             "Karyorelict Nuclear",
                             "Condylostoma Nuclear",
                             "Mesodinium Nuclear", "Peritrich Nuclear",
                             "Blastocrithidia Nuclear"])
    })

    max_objects: int = field(default=10000000, metadata={
                "metadata": {
                    "description": "Maximum number of objects to keep in memory while loading data into the database. \
Linearly related to memory usage and inversely correlated with runtime."},
                "validate": validate.Range(min=1)
    })
    max_regression: float = field(default=0.2, metadata={
                "metadata": {
                    "description": "if the ORF lacks a valid start site, this percentage indicates how far along \
the sequence Mikado should look for a good start site. Eg. with a value of 0.1, on a 300bp sequence with an open ORF \
Mikado would look for an alternative in-frame start codon in the first 30 bps (10% of the cDNA)."},
            "validate": validate.Range(min=0.0, max=1.0),
            "required": True
    })
    start_adjustment: bool = field(default=True, metadata={
                "metadata": {"description": "Boolean switch. If set to true (default), if an ORF is truncated at \
the 5' Mikado will look for internal start codons. See 'max_regression'."},
    })
    max_target_seqs: int = field(default=100000, metadata={
                "metadata": {
                    "description": "Equivalently to BLAST, it indicates the maximum number of targets \
to keep per blasted sequence."},
                "validate": validate.Range(min=1)
    })
    force: bool = field(default=True, metadata={
                "metadata": {"description": "Whether to drop and reload everything into the DB"},
    })
    single_thread: bool = field(default=False, metadata={})
    shm: bool = field(default=False, metadata={
        "metadata": {"description": "Whether to build the database in /dev/shm before copying it to its final "
                                    "location, or not."}
    })



import functools

from marshmallow_dataclass import dataclass, List, Dict, Union, Optional
from dataclasses import field, asdict
from marshmallow import validate
from .transcript_base import TranscriptBase
available_metrics = TranscriptBase.get_available_metrics()
from ..exceptions import InvalidConfiguration
import re


key_pattern = re.compile(r"([^ ()]+)")


# DO NOT TOUCH THIS. The compilation at each call, no lazy storing, is necessary to prevent
# multiprocessing madness (eval and pickle do not mix at all). The LRU cache hopefully speeds things up
# in the context of a single process
@functools.lru_cache(maxsize=1000, typed=True)
def compiler(expression):
    try:
        return compile(expression, "<json>", "eval")
    except SyntaxError:
        raise InvalidConfiguration("Invalid expression:\n{}".format(expression))


class Unique(validate.Validator):
    message = "{input} is Not unique"

    def __init__(self, error=None):
        self.error = error

    def _repr_args(self):
        return ""

    def _format_error(self, value):
        return self.message.format(input=value)

    def __call__(self, value):
        if len(value) != len(set(value)):
            raise validate.ValidationError(self._format_error(value))
        return value


@dataclass
class SizeFilter:
    operator: str = field(metadata={"required": True, "validate": validate.OneOf(["gt", "ge", "lt", "le"])})
    value: float = field(metadata={"required": True})
    metric: Optional[str] = field(metadata={"required": False}, default=None)
    name: Optional[str] = field(default=None)
    source: Optional[str] = field(default=None)
    default: Optional[float] = field(default=0)


@dataclass
class NumBoolEqualityFilter:
    value: Union[float, bool] = field(metadata={"required": True})
    operator: str = field(metadata={"required": True, "validate": validate.OneOf(["ne", "eq"])})
    metric: Optional[str] = field(metadata={"required": False}, default=None)
    name: Optional[str] = field(default=None)
    source: Optional[str] = field(default=None)
    default: Optional[Union[float, bool]] = field(default=0)


@dataclass
class InclusionFilter:
    value: list = field(metadata={"required": True, "validate": [Unique]})
    operator: str = field(metadata={"required": True, "validate": validate.OneOf(["in", "not in"])})
    metric: Optional[str] = field(metadata={"required": False}, default=None)
    name: Optional[str] = field(default=None)
    source: Optional[str] = field(default=None)
    default: Optional[List[float]] = field(default=None)


@dataclass
class RangeFilter:
    value: List[float] = field(metadata={
        "required": True,
        "validate": [validate.Length(min=2, max=2), Unique]})
    operator: str = field(metadata={"required": True, "validate": validate.OneOf(["within", "not within"])})
    metric: Optional[str] = field(metadata={"required": False}, default=None)
    name: Optional[str] = field(default=None)
    default: Optional[float] = field(default=0)


@dataclass
class MinMaxScore:
    rescaling: str = field(metadata={"required": True, "validate": validate.OneOf(["max", "min"])})
    filter: Optional[Union[SizeFilter, NumBoolEqualityFilter, RangeFilter, InclusionFilter]]
    use_raw: bool = field(default=False)
    multiplier: float = field(default=1, metadata={"required": False})
    default: Optional[Union[float, bool]] = field(default=0, metadata={"required": False})
    rtype: Optional[str] = field(default="float", metadata={"validate": validate.OneOf(["float", "int", "bool"]),
                                                            "required": False})
    percentage: Optional[bool] = field(default=False)


@dataclass
class TargetScore:
    rescaling: str = field(metadata={"required": True, "validate": validate.OneOf(["target"])})
    value: Union[float, bool] = field(metadata={"required": True})
    filter: Optional[Union[SizeFilter, NumBoolEqualityFilter, RangeFilter, InclusionFilter]]
    # Use_raw must be false for target scores
    use_raw: bool = field(default=False, metadata={"validate": validate.OneOf([False])})
    multiplier: float = field(default=1, metadata={"required": False})
    default: Optional[Union[float, bool]] = field(default=0, metadata={"required": False})
    rtype: Optional[str] = field(default="float", metadata={"validate": validate.OneOf(["float", "int", "bool"]),
                                                            "required": False})
    percentage: Optional[bool] = field(default=False)


@dataclass
class Requirements:
    parameters: Dict[str, Union[SizeFilter, InclusionFilter, NumBoolEqualityFilter, RangeFilter]] = field(
        metadata={"validate": validate.Length(min=1), "required": True}
    )
    expression: List[str] = field(default_factory=lambda: [])

    _expression = None

    @staticmethod
    def _create_expression(expression: list, parameters: dict):
        if len(expression) == 0:
            expression = " and ".join(list(parameters.keys()))
            keys = parameters.keys()
        else:
            expression = " ".join(expression)
            keys = set([key for key in key_pattern.findall(expression) if key not in ("and", "or", "not", "xor")])
            diff_params = set.difference(set(keys), set(parameters.keys()))
            if len(diff_params) > 0:
                raise InvalidConfiguration("Expression and required parameters mismatch:\n\t{0}".format(
                    "\n\t".join(list(diff_params))))
        for key in keys:  # Create the final expression
            expression = re.sub(r"\b{}\b".format(key), "evaluated[\"{0}\"]".format(key), expression)
        return expression

    def _check_parameters(self):
        # Check that the parameters are valid
        parameters_not_found = []

        for key in self.parameters:
            key_value = None
            dots = key.split(".")
            if dots[0] == "external" or dots[0] == "attributes":
                if len(dots) == 1 or len(dots) > 3:
                    parameters_not_found.append(dots[0])
                    continue
                elif len(dots) == 2:
                    key_name = ".".join(dots)
                else:
                    key_name = ".".join(dots[:-1])
                    # print(key_name)
                key_value = dots[1]
            else:
                key_name = dots[0]
            if key_name not in available_metrics:
                if key_value is not None:
                    pass
                else:
                    parameters_not_found.append(key_name)
                    continue
            self.parameters[key].name = key_name
            if key_value:
                self.parameters[key].source = key_value

        if len(parameters_not_found) > 0:
            raise InvalidConfiguration(
                "The following parameters, selected for filtering, are invalid:\n\t{0}".format(
                    "\n\t".join(parameters_not_found)
                ))

    # If the requirements are changed at run time, this *needs* to be called again
    def _check_my_requirements(self):
        """Check that the provided parameters are as expected"""
        self._expression = None
        assert hasattr(self, "parameters"), (type(self), self.__dict__)
        self._check_parameters()
        self._expression = self._create_expression(self.expression, self.parameters)
        _ = self.compiled

    @property
    def compiled(self):
        return compiler(self._expression)

    def copy(self):
        return self.Schema().load(asdict(self))


@dataclass
class ScoringFile:
    scoring: Dict[str, Union[TargetScore, MinMaxScore]] = field(metadata={"validate": validate.Length(min=1)})
    requirements: Requirements
    not_fragmentary: Optional[Requirements] = field(default=None)
    as_requirements: Optional[Requirements] = field(default=None)
    cds_requirements: Optional[Requirements] = field(default=None)

    def _create_missing_reqs(self, minimal_orf_length):
        """"""

        assert self.requirements.parameters

        for section_name in ["not_fragmentary", "as_requirements", "cds_requirements"]:
            section = getattr(self, section_name)
            if section is not None:
                continue
            if section_name == "cds_requirements":
                # If no CDS requirement section is present, presume that the only req. is that the ORF
                # is equal or longer than the "minimal_orf_length" (taken from pick.orf_loading)
                mock = {
                    "parameters": {"selected_cds_length": {
                        "operator": "ge",
                        "value": minimal_orf_length}},
                    "expression": ["selected_cds_length"]
                }
                self.cds_requirements = Requirements.Schema().load(mock)
                assert self.cds_requirements.parameters
            else:
                setattr(self, section_name, self.requirements.copy())

    def _check_scoring(self):
        """Method to check that the specified metrics are correct"""
        parameters_found = set()
        parameters_not_found = []
        double_parameters = []
        invalid_raw = set()
        invalid_filter = set()

        for parameter in self.scoring:
            if parameter not in available_metrics:
                # Leniency for external_scores
                if "." in parameter and len(parameter.split(".")) == 2 and (parameter.split(".")[0] == "external" or
                                                                            parameter.split(".")[0] == "attributes"):
                    pass
                else:
                    parameters_not_found.append(parameter)

            if parameter in parameters_found:
                double_parameters.append(parameter)

            if (not parameter.startswith("external") and not parameter.startswith("attributes") and
                    getattr(TranscriptBase, parameter).usable_raw is False and
                    self.scoring[parameter].use_raw is True):
                invalid_raw.add(parameter)

            if self.scoring[parameter].filter:
                metric = self.scoring[parameter].filter.metric
                if metric is not None and metric != "" and metric not in available_metrics:
                    parameters_not_found.append(metric)

        if len(parameters_not_found) > 0 or len(double_parameters) > 0 or len(invalid_filter) > 0 or len(
                invalid_raw) > 0:
            err_message = ''
            if len(parameters_not_found) > 0:
                err_message = """The following parameters, present in the JSON file,
                    are not available!\n\t{0}\n""".format(
                    "\n\t".join(parameters_not_found))
            if len(double_parameters) > 0:
                err_message += """The following parameters have been specified more than once,
                    please correct:\n\t{0}""".format("\n\t".join(list(double_parameters)))
            if len(invalid_filter) > 0:
                err_message += """The following parameters have an invalid filter,
                    please correct:
                    \t{0}""".format("\n\t".join(list(invalid_filter)))
            if len(invalid_raw) > 0:
                err_message += """The following parameters cannot be used as raw scores, either
                    because they are not normalized on their own, or because a "target" rescaling has been asked for:
                    \t{0}""".format("\n\t".join(list(invalid_raw)))
            raise InvalidConfiguration(err_message)

    def check(self, minimal_orf_length):
        self.requirements._check_my_requirements()
        self._create_missing_reqs(minimal_orf_length=minimal_orf_length)
        for name, section in zip(["as_requirements", "cds_requirements", "not_fragmentary"],
                                            [self.as_requirements, self.cds_requirements, self.not_fragmentary]):
            assert section is not None, name
            section._check_my_requirements()
        self._check_scoring()



#!/usr/bin/env python3

"""
Module that implements the Reid/Daid/Hurley algorithm for community finding.
"""

import networkx
from ..utilities.log_utils import create_null_logger
from collections import defaultdict
from itertools import chain, combinations


def find_communities(graph: networkx.Graph, logger=None) -> set:
    """

    :param graph: a Graph instance from networkx
    :type graph: networkx.Graph

    :param logger: optional logger. A default null one will be created if none is provided.
    :type logger: (None|logging.Logger)

    This function is a wrapper around the networkX methods to find communities inside a graph.
    The method takes as input a precomputed graph and returns a set of the available communities.
    """
    if logger is None:
        logger = create_null_logger()

    logger.debug("Creating the communities for %s", logger.name)
    communities = [frozenset(comm) for comm in networkx.connected_components(graph)]

    logger.debug("Communities for %s:\n\t\t%s", logger.name, "\n\t\t".join(
        [str(_) for _ in communities]))
    # result = [frozenset(x) for x in communities.values()]
    for element in set.difference(set(graph.nodes()), set(chain(*communities[:]))):
        communities.append(frozenset([element]))
    return set(communities)


def define_graph(objects: dict, inters, **kwargs) -> networkx.Graph:
    """
    :param objects: a dictionary of objects to be grouped into a graph
    :type objects: dict

    :param inters: the intersecting function to be used to define the graph
    :type inters: callable

    :param kwargs: optional arguments to be passed to the inters function
    :type kwargs: dict

    This function will compute the graph which will later be used by find_communities.
    The method takes as mandatory inputs the following:
        - "objects" a dictionary of objects that form the graph
        - "inters" a function/method that determines whether two objects are connected or not.

    It will then return a graph.
    The method accepts also kwargs that can be passed to the inters function.
    WARNING: the kwargs option is really stupid and does not check
    for correctness of the arguments!
    """

    graph = networkx.Graph()

    # As we are using intern for transcripts, this should prevent
    # memory usage to increase too much
    graph.add_nodes_from(objects.keys())

    for obj, other_obj in combinations(objects.keys(), 2):
        if obj == other_obj:
            continue
        elif inters(objects[obj], objects[other_obj], **kwargs):
            # Connections are not directional
            graph.add_edge(*tuple(sorted([obj, other_obj])))

    return graph


def find_cliques(graph: networkx.Graph, logger=None) -> (networkx.Graph, list):
    """

    :param graph: graph to which it is necessary to call the cliques for.

    :param logger: optional logger for the function

    Wrapper for the BronKerbosch algorithm, which returns the maximal cliques in the graph.
    It is the new interface for the BronKerbosch function, which is not called directly
    from outside this class any longer.
    The "inters" keyword provides the function used to determine
    whether two vertices are connected or not in the graph.
    """

    if logger is None:
        logger = create_null_logger()

    logger.debug("Creating cliques for %s", logger.name)
    cliques = [frozenset(x) for x in networkx.find_cliques_recursive(graph)]
    logger.debug("Created %d cliques for %s", len(cliques), logger.name)

    return cliques



"""
XML serialisation class.
"""

import os
import logging.handlers as logging_handlers
import logging
from sqlalchemy import inspect
from sqlalchemy.orm.session import Session
from ...utilities.dbutils import DBBASE
import pysam
from ...utilities.dbutils import connect
from ...utilities.log_utils import create_null_logger, check_logger
from ...configuration import MikadoConfiguration, DaijinConfiguration
from .query import Query, id_pattern
from .target import Target
from .hsp import Hsp
from .hit import Hit
from .utils import load_into_db
from .xml_utils import get_multipliers
from .xml_serialiser import _serialise_xmls
from .tab_serialiser import _serialise_tabular
import pandas as pd
import multiprocessing


__author__ = 'Luca Venturini'


# A serialisation class must have a ton of attributes ...
# pylint: disable=too-many-instance-attributes


class BlastSerializer:
    """This class has the role of taking in input a blast XML file and (partially)
    serialise it into a database. We are using SQLalchemy, so the database type
    could be any of SQLite, MySQL, PSQL, etc."""

    __name__ = "BlastSerializer"
    logger = create_null_logger(__name__)

    # This evaluates to characters A-z and |. Used to detect true matches
    # in the match line of blast alignments
    # Put here so it's calculated *once* inside the program
    __valid_matches = set([chr(x) for x in range(65, 91)] +
                          [chr(x) for x in range(97, 123)] +
                          ["|", "*"])

    def __init__(self, xml_name,
                 logger=None,
                 configuration=None):
        """Initializing method. Arguments:

        :param xml_name: The XML(s) to parse.

        Arguments:

        :param configuration: a configuration dictionary.
        :type configuration: (MikadoConfiguration|DaijinConfiguration)


        """

        if configuration is None:
            raise ValueError("No configuration provided!")

        if logger is not None:
            self.logger = check_logger(logger)
        else:
            raise ValueError("No logger provided!")
        self.logger.debug("Started to serialise %s, log level: %s",
                         xml_name, self.logger.level)

        # Runtime arguments

        self.procs = configuration.threads
        self.single_thread = configuration.serialise.single_thread
        self.configuration = configuration
        # pylint: disable=unexpected-argument,E1123
        multiprocessing.set_start_method(self.configuration.multiprocessing_method, force=True)
        # pylint: enable=unexpected-argument,E1123
        self.logger.info("Number of dedicated workers: %d", self.procs)
        self.logging_queue = multiprocessing.Queue(-1)
        self.logger_queue_handler = logging_handlers.QueueHandler(self.logging_queue)
        self.queue_logger = logging.getLogger("parser")
        self.queue_logger.addHandler(self.logger_queue_handler)
        self.queue_logger.setLevel(self.configuration.log_settings.log_level)
        self.queue_logger.propagate = False
        self.log_writer = logging_handlers.QueueListener(self.logging_queue, self.logger)
        self.log_writer.start()

        self._max_target_seqs = configuration.serialise.max_target_seqs
        self.maxobjects = configuration.serialise.max_objects
        target_seqs = configuration.serialise.files.blast_targets
        query_seqs = configuration.serialise.files.transcripts

        self.header = None
        if xml_name is None:
            self.logger.warning("No BLAST XML provided. Exiting.")
            return

        self.engine = connect(configuration)

        self._blast_loading_debug = self.configuration.serialise.files.blast_loading_debug
        if self._blast_loading_debug:
            self.logger.warning("Activating the XML debug mode")
            self.single_thread = True
            self.procs = 1

        DBBASE.metadata.create_all(self.engine)  # @UndefinedVariable
        session = Session(bind=self.engine)
        self.session = session
        self.hit_i_string = str(Hit.__table__.insert(bind=self.engine).compile())
        self.hsp_i_string = str(Hsp.__table__.insert(bind=self.engine).compile())
        # Remove indices
        self.logger.debug("Created the session")
        # Load sequences if necessary
        self.__determine_sequences(query_seqs, target_seqs)
        self.xml = xml_name
        # Just a mock definition
        self.not_pickable = ["manager", "printer_process",
                             "context", "logger_queue_handler", "queue_logger",
                             "log_writer",
                             "log_process", "pool", "main_logger",
                             "log_handler", "log_writer", "logger", "session",
                             "get_query", "engine", "query_seqs", "target_seqs"]

        self.queries, self.targets = dict(), dict()

        self.logger.debug("Finished __init__")

    def __getstate__(self):

        state = self.__dict__.copy()
        for not_pickable in self.not_pickable:
            if not_pickable in state:
                del state[not_pickable]

        return state

    def __determine_sequences(self, query_seqs, target_seqs):

        """Private method to assign the sequence file variables
        if necessary.
        :param query_seqs:
        :param target_seqs:
        :return:
        """

        if isinstance(query_seqs, str):
            assert os.path.exists(query_seqs), "Invalid FASTA file name, as it does not exist: {}".format(query_seqs)
            self.query_seqs = pysam.FastaFile(query_seqs)
        elif query_seqs is None:
            self.query_seqs = None
        else:
            self.logger.warn("Query type: %s", type(query_seqs))
            self.query_seqs = query_seqs

        self.target_seqs = []
        for target in target_seqs:
            if not os.path.exists(target):
                raise ValueError("{} not found!".format(target))
            self.target_seqs.append(pysam.FastaFile(target))

        return

    def __serialize_queries(self, queries):

        """Private method used to serialise the queries.
        Additional argument: a set containing all queries already present
        in the database."""

        counter = 0
        self.logger.info("Started to serialise the queries")
        objects = []
        for record, length in zip(self.query_seqs.references, self.query_seqs.lengths):
            if not record:
                continue

            if id_pattern.search(record):
                record = id_pattern.search(record).groups()[0]

            if record in queries and queries[record][1] == length:
                continue
            elif record in queries:
                if queries[record][1] == length:
                    continue
                else:
                    raise KeyError("Discrepant length for query {} (original {}, new {})".format(
                        record, queries[record][1], length))
                # self.session.query(Query).filter(Query.query_name == record).update(
                #     {"query_length": length})
                # queries[record] = (queries[record][0], length)

            objects.append({
                "query_name": record,
                "query_length": length
            })

            if len(objects) >= self.maxobjects:
                self.logger.debug("Loading %d objects into the \"query\" table (total %d)",
                                 self.maxobjects, counter)

                # pylint: disable=no-member
                self.session.begin(subtransactions=True)
                # self.session.bulk_insert_mappings(Query, objects)
                self.engine.execute(Query.__table__.insert(), objects)
                # pylint: enable=no-member

                self.session.commit()
                counter += len(objects)
                objects = []

        if len(objects) > 0:
            self.logger.debug("Loading %d objects into the \"query\" table (total %d)",
                             len(objects), counter+len(objects))
            # pylint: disable=no-member
            counter += len(objects)
            # pylint: disable=no-member
            self.session.begin(subtransactions=True)
            self.engine.execute(Query.__table__.insert(), objects)
            # pylint: enable=no-member
            self.session.commit()
            # pylint: enable=no-member

        self.logger.info("Loaded %d objects into the \"query\" table", counter)

        if self.single_thread is True or self.procs == 1:
            _queries = pd.read_sql_table("query", self.engine, index_col="query_name",
                                        columns=["query_id", "query_length"])
            queries = dict((qname, int(qid)) for
                            qname, qid in zip(_queries.index, _queries["query_id"].values))
            self.logger.info("%d in queries", len(queries))
        else:
            queries = dict()

        return queries

    def __serialize_targets(self, targets):
        """
        This private method serialises all targets contained inside the target_seqs
        file into the database.
        :param targets: a cache dictionary which records whether the sequence
          is already present in the DB or not.
        """

        counter = 0
        objects = []
        self.logger.info("Started to serialise the targets")
        for target in self.target_seqs:
            for record, length in zip(target.references, target.lengths):
                if id_pattern.search(record):
                    record = id_pattern.search(record).groups()[0]

                if record in targets and targets[record][1] is True:
                    continue
                elif record in targets:
                    self.session.query(Target).filter(Target.target_name == record).update(
                        {"target_length": length})
                    targets[record] = (targets[record][0], True)
                    continue

                objects.append({
                    "target_name": record,
                    "target_length": length
                })
                counter += 1
                if len(objects) >= self.maxobjects:
                    self.logger.debug("Loading %d objects into the \"target\" table",
                                     counter)
                    self.session.begin(subtransactions=True)
                    self.engine.execute(Target.__table__.insert(), objects)
                    self.session.commit()
                    objects = []
        self.logger.debug("Loading %d objects into the \"target\" table, (total %d)",
                         len(objects), counter)
        # pylint: disable=no-member
        self.session.begin(subtransactions=True)
        self.engine.execute(Target.__table__.insert(), objects)
        # pylint: enable=no-member
        self.session.commit()

        self.logger.info("Loaded %d objects into the \"target\" table", counter)
        if self.single_thread is True or self.procs == 1:
            _targets = pd.read_sql_table("target", self.engine, index_col="target_name",
                                         columns=["target_id", "target_length"])
            targets = dict((qname, int(qid)) for
                            qname, qid in zip(_targets.index, _targets["target_id"].values))
            self.logger.debug("%d in targets", len(targets))
        else:
            targets = dict()
        return targets

    def __serialise_sequences(self):

        """ Private method called at the beginning of serialize. It is tasked
        with loading all necessary FASTA sequences into the DB and precaching the IDs.
        """

        targets = dict()
        queries = dict()
        self.logger.debug("Loading previous IDs")
        for query in self.session.query(Query):
            queries[query.query_name] = (query.query_id, query.query_length)
        for target in self.session.query(Target):
            targets[target.target_name] = (target.target_id, target.target_length)
        self.logger.debug("Loaded previous IDs; %d for queries, %d for targets",
                         len(queries), len(targets))

        self.logger.debug("Started the sequence serialisation")
        if self.target_seqs:
            targets = self.__serialize_targets(targets)
        if self.query_seqs is not None:
            queries = self.__serialize_queries(queries)

        return queries, targets

    def _load_into_db(self, hits, hsps, force=False):
        """

        :param hits:
        :param hsps:

        :param force: boolean flag. If set, data will be loaded no matter what.
        To be used at the end of the serialisation to load the final batch of data.
        :type force: bool

        :return:
        """
        load_into_db(self, hits, hsps, force=force)

    def serialize(self):

        """Method to serialize the BLAST XML file into a database
        provided with the __init__ method """
        self.logger.debug("Checking the sequences")
        self.queries, self.targets = self.__serialise_sequences()
        self.logger.debug("Checked the sequences")

        if isinstance(self.xml, str):
            if ".xml" in self.xml:
                self.__serialise_xmls()
            else:
                self.__serialise_tabular()
        else:
            assert isinstance(self.xml, (list, set))
            if all([".xml" in fname for fname in self.xml]):
                self.__serialise_xmls()
            else:
                self.__serialise_tabular()

    def __serialise_tabular(self):
        """Parser to perform the analysis of tabular BLAST files."""
        _serialise_tabular(self)

    def __serialise_xmls(self):
        _serialise_xmls(self)

    def __call__(self):
        """
        Alias for serialize
        """
        inspector = inspect(self.engine)
        hit_indices = dict((idx.name, idx) for idx in Hit.__table__.indexes)
        hsp_indices = dict((idx.name, idx) for idx in Hsp.__table__.indexes)

        # More robust: only drop existing indices
        for idx in inspector.get_indexes(Hit.__table__.name):
            index = hit_indices[idx["name"]]
            index.drop(bind=self.engine)
        for idx in inspector.get_indexes(Hsp.__table__.name):
            index = hsp_indices[idx["name"]]
            index.drop(bind=self.engine)

        self.engine.execute("PRAGMA foreign_keys=OFF")
        self.serialize()
        # Recreate the indices
        [idx.create(bind=self.engine) for idx in Hit.__table__.indexes]
        [idx.create(bind=self.engine) for idx in Hsp.__table__.indexes]
        self.engine.execute("PRAGMA foreign_keys=ON")

# pylint: enable=too-many-instance-attributes

    @staticmethod
    def get_multipliers(record, application=None):
        return get_multipliers(record, application=application)



#!/usr/bin/env python3
import argparse
import sys
from ...exceptions import InvalidTranscript, InvalidParsingFormat
from ...parsers import parser_factory
from ...parsers.GFF import GFF3
from ...parsers.GTF import GTF
from ...parsers.bed12 import BED12, Bed12Parser
from ...loci import Transcript, Gene


def _convert_bam(parser, args, out_format):
    current = None
    mock_gene_counter = 0
    _line_counter = 0
    for line in parser:
        if line.is_unmapped is True:
            continue

        mock_gene_counter += 1
        gene = "gene_{mock_gene_counter}".format(**locals())
        try:
            transcript = Transcript(line)
        except KeyboardInterrupt:
            raise
        except Exception as exc:
            raise InvalidTranscript("Invalid BAM line at position {} for file {}. Error: {}".format(
                _line_counter, parser.name, exc
            ))
        transcript.parent = gene
        print(Gene(transcript).format(out_format, transcriptomic=args.transcriptomic), file=args.out)
    return


def _convert_gtf(parser: GTF, args, out_format):
    genes = dict()
    current = None
    found_ids = set()

    for line_counter, line in enumerate(parser, start=1):
        if line.header is True:
            continue
        elif hasattr(line, "feature") and "superlocus" in line.feature or line.feature in ("chromosome", "region"):
            continue
        if line.gene in genes:
            if line.is_transcript is True and line.transcript in genes[line.gene]:
                if args.assume_sorted is True:
                    raise InvalidParsingFormat(
                        f"The transcript at line no. {line_counter} has already appeared: the input is not sorted.")
                gene = genes[line.gene]
                assert isinstance(gene, Gene)
                assert line.strand == gene.strand
                assert line.chrom == gene.chrom
                gene.transcripts[line.transcript].start, gene.transcripts[line.transcript].end = line.start, line.end
                gene.transcripts[line.transcript]._set_expandable(False)
                gene.transcripts[line.transcript].attributes.update(line.attributes)
            elif line.is_transcript is True:
                if args.assume_sorted is True and line.transcript in found_ids:
                    raise InvalidParsingFormat(
                        f"The transcript at line no. {line_counter} has already appeared: the input is not sorted.")
                genes[line.gene].add(Transcript(line))
            elif line.is_gene:
                if line.gene in genes or line.gene in found_ids:
                    if args.assume_sorted is True:
                        raise InvalidParsingFormat(
                            f"The transcript at line no. {line_counter} has already appeared: the input is not sorted.")
                genes[line.gene].start, genes[line.gene].end = line.start, line.end
                genes[line.gene].attributes.update(line.attributes)
            elif line.is_exon:
                if line.transcript in genes[line.gene].transcripts:
                    genes[line.gene].add_exon(line)
                else:
                    genes[line.gene].add(Transcript(line))
        else:
            if line.is_gene:
                if args.assume_sorted is True and line.gene in genes:
                    raise InvalidParsingFormat(
                        f"The gene at line no. {line_counter} has already appeared: the input is not sorted.")
                genes[line.gene] = Gene(line)
            else:
                if args.assume_sorted is True and line.gene in found_ids or line.transcript in found_ids:
                    raise InvalidParsingFormat(
                        f"The gene at line no. {line_counter} has already appeared: the input is not sorted.")
                genes[line.gene] = Gene(Transcript(line))

        found_ids.add(line.transcript)
        if current and current != line.gene and args.assume_sorted is True:
            print(genes[current].format(out_format, transcriptomic=args.transcriptomic), file=args.out)
            del genes[current]

        current = line.gene

    if args.assume_sorted is False:
        for gid, gene in genes.items():
            print(gene.format(out_format, transcriptomic=args.transcriptomic), file=args.out)
    elif current is not None:
        print(genes[current].format(out_format, transcriptomic=args.transcriptomic), file=args.out)

    return


def _convert_bed12(parser: Bed12Parser, args, out_format):
    mock_gene_counter = 0
    for line in parser:
        if line.header is True:
            continue
        gid = "gene_{mock_gene_counter}".format(**locals())
        line.parent = gid
        gene = Gene(Transcript(line))
        print(gene.format(out_format, transcriptomic=args.transcriptomic), file=args.out)

    return


def _convert_gff(parser: GFF3, args, out_format):
    orphaned = dict()
    tid2gene = dict()
    genes = dict()

    current = None
    for line in parser:
        if line.header is True:
            continue
        elif line.feature in ("region", "chrom", "superlocus", "protein"):
            continue
        # We need to have better checks here regarding sorted/unsorted inputs here
        if line.is_gene is False and args.assume_sorted is True:
            if line.is_transcript is False:
                if current is None or line.parent[0] not in genes[current].transcripts:
                    raise InvalidParsingFormat("The provided GFF is not sorted.")
                genes[current].add_exon(line)
            else:
                assert genes[current].id in line.parent[0]
                genes[current].add(Transcript(line))
        elif line.is_gene is True:
            if args.assume_sorted is True and current is not None:
                print(genes[current].format(out_format, transcriptomic=args.transcriptomic), file=args.out)
                del genes[current]
            genes[line.gene] = Gene(line)
            current = line.gene
            if args.assume_sorted is False:
                orph = list(orphaned.keys())
                for tid in orph:
                    if orphaned[tid].parent and orphaned[tid].parent[0] == current:
                        genes[current].add(orphaned[tid])
                        del orphaned[tid]
        elif line.is_transcript is True and args.assume_sorted is False:
            transcript = Transcript(line)
            if transcript.parent:
                tid2gene[transcript.id] = transcript.parent[0]
            if transcript.id in orphaned:
                assert transcript.chrom == orphaned[transcript.id].chrom
                assert transcript.strand == orphaned[transcript.id].strand
                orphaned[transcript.id].start, orphaned[transcript.id].end = transcript.start, transcript.end
                orphaned[transcript.id].attributes.update(transcript.attributes)
                orphaned[transcript.id]._set_expandable(False)
                orphaned[transcript.id].parent = transcript.gene
                if orphaned[transcript.id].parent and orphaned[transcript.id].parent[0] in genes:
                    genes[orphaned[transcript.id].parent[0]].add(orphaned[transcript.id])
                    del orphaned[transcript.id]
            elif transcript.parent and transcript.parent[0] in genes:
                genes[transcript.parent[0]].add(transcript)
            else:
                orphaned[transcript.id] = transcript
        else:
            assert args.assume_sorted is False
            # Exons might have multiple parents ...
            if any(_ in tid2gene for _ in line.parent):
                gids = set([tid2gene[_] for _ in line.parent])
                assert len(gids) == 1, line
                gid = gids.pop()
                if gid in genes:
                    genes[gid].add_exon(line)
                    for parent in line.parent:
                        if parent not in genes[gid]:
                            genes[gid].add(Transcript(line))
                else:
                    for parent in line.parent:
                        if parent in orphaned:
                            orphaned[parent].add_exon(line)
                        else:
                            orphaned[parent] = Transcript(line)
            else:
                for parent in line.parent:
                    if parent in orphaned:
                        orphaned[parent].add_exon(line)
                    else:
                        orphaned[parent] = Transcript(line)

    if current and args.assume_sorted is True:
        print(genes[current].format(out_format, transcriptomic=args.transcriptomic), file=args.out)

    if args.assume_sorted is False:
        for tid in orphaned:
            if not orphaned[tid].parent:
                orphaned[tid].parent = "{}.gene".format(orphaned[tid].id)
                genes[orphaned[tid].parent[0]] = orphaned[tid]
            else:
                for parent in orphaned[tid].parent:
                    if parent in genes:
                        genes[parent].add(orphaned[tid])
                    else:
                        genes[parent] = Gene(orphaned[tid])

        for gid, gene in genes.items():
            print(gene.format(out_format, transcriptomic=args.transcriptomic), file=args.out)

    return


def launch(args):

    if args.gf == "-":
        if args.in_format is None:
            raise ValueError("I need a format if it cannot be inferred from the string")
        parser = parser_factory(sys.stdin, input_format=args.in_format)
    else:
        parser = parser_factory(args.gf, input_format=args.in_format)

    if parser.__annot_type__ == "gtf":
        out_format = "gff3"
    elif parser.__annot_type__ in ("bed12", "gff3", "bam"):
        out_format = "gtf"
    else:
        raise TypeError("Invalid annotation type: {}".format(parser.__annot_type__))

    if args.out_format is not None and args.out_format != out_format:
        if args.out_format != parser.__annot_type__ or args.transcriptomic is True:
            out_format = args.out_format
        else:
            print("Input and output format are the same, aborting.", file=sys.stderr)
            sys.exit(1)

    if out_format == "gff3":
        print("##gff-version\t3", file=args.out)

    if parser.__annot_type__ == "bam":
        _convert_bam(parser, args, out_format)
    elif parser.__annot_type__ == "gtf":
        _convert_gtf(parser, args, out_format)
    elif parser.__annot_type__ == "bed12":
        _convert_bed12(parser, args, out_format)
    else:
        _convert_gff(parser, args, out_format)

    if not args.out == sys.stdout:
        args.out.flush()
        args.out.close()


def convert_parser():
    """
    Parser for the command line
    :return:
    """

    parser = argparse.ArgumentParser(
        "Utility to covert across GTF, GFF3 and BED12.")
    parser.add_argument("-as", "--assume-sorted", default=False, action="store_true")
    parser.add_argument("-of", "--out-format", dest="out_format",
                        choices=["bed12", "gtf", "gff3"], default=None)
    parser.add_argument("-if", "--in-format", dest="in_format",
                        choices=["bed12", "gtf", "gff3", "bam"], default=None)
    parser.add_argument("-t", "--transcriptomic", default=False, action="store_true",
                        help="Flag. If on, the file will be converted to a transcriptomic version.")
    parser.add_argument("gf")
    parser.add_argument("out", type=argparse.FileType("w"), nargs="?", default=sys.stdout)
    parser.set_defaults(func=launch)
    return parser



#!/usr/bin/env python3
# coding: utf-8

"""
Subprogram that constitutes the first step of the Mikado pipeline.
"""

import sys
import os
import argparse
import logging
import logging.handlers

from Mikado import version

from ._utils import check_log_settings_and_create_logger
from ..configuration import MikadoConfiguration, DaijinConfiguration, parse_list_file
from ..configuration.configurator import load_and_validate_config
from ..exceptions import InvalidConfiguration
from collections import Counter
from typing import Union
from ..preparation.prepare import prepare


__author__ = 'Luca Venturini'


def parse_gff_args(mikado_config, args):
    __gff_counter = Counter()
    __gff_counter.update(args.gff)
    if __gff_counter.most_common()[0][1] > 1:
        raise InvalidConfiguration(
            "Repeated elements among the input GFFs! Duplicated files: {}".format(
                ", ".join(_[0] for _ in __gff_counter.most_common() if _[1] > 1)
            ))
    mikado_config.prepare.files.gff = args.gff
    num_files = len(mikado_config.prepare.files.gff)
    if args.strand_specific:
        mikado_config.prepare.strand_specific = True
    elif args.strand_specific_assemblies:
        strand_specific_assemblies = args.strand_specific_assemblies.split(",")
        if len(strand_specific_assemblies) > num_files:
            raise InvalidConfiguration("Incorrect number of strand-specific assemblies specified!")
        for member in strand_specific_assemblies:
            if member not in mikado_config.prepare.files.gff:
                raise InvalidConfiguration("Incorrect assembly file specified as strand-specific")
        mikado_config.prepare.files.strand_specific_assemblies = strand_specific_assemblies
    if args.labels:
        labels = args.labels.split(",")
        # Checks labels are unique
        if len(set(labels)) < len(labels):
            raise InvalidConfiguration("Duplicated labels detected!")
        elif any([True for _ in labels if _.strip() == '']):
            raise InvalidConfiguration("Empty labels provided!")
        elif len(labels) != num_files:
            raise InvalidConfiguration("Incorrect number of labels specified")
        mikado_config.prepare.files.labels = labels
    else:
        if not mikado_config.prepare.files.labels:
            labels = [str(_) for _ in list(range(1, 1 + num_files))]
            mikado_config.prepare.files.labels = labels
    mikado_config.prepare.files.exclude_redundant = [False] * len(mikado_config.prepare.files.gff)
    mikado_config.prepare.files.reference = [False] * len(mikado_config.prepare.files.gff)
    return mikado_config


def parse_prepare_options(args, mikado_config) -> Union[DaijinConfiguration, MikadoConfiguration]:

    if args.reference is not None:
        if hasattr(args.reference, "close") and hasattr(args.reference, "name"):
            args.reference.close()
            mikado_config.reference.genome = args.reference.name
        elif hasattr(args.reference, "close") and hasattr(args.reference, "filename"):
            # Pysam FastaFile. The filename is bytes, not str
            args.reference.close()
            mikado_config.reference.genome = args.reference.filename.decode()
        elif isinstance(args.reference, bytes):
            mikado_config.reference.genome = args.reference.decode()
        elif isinstance(args.reference, str):
            mikado_config.reference.genome = args.reference
        else:
            raise InvalidConfiguration(f"Invalid value type for the reference: {args.reference} (type "
                                       f"{type(args.reference)}")

    if not os.path.exists(mikado_config.reference.genome):
        raise InvalidConfiguration("Reference genome file {} is not available. Please double check.".format(
            mikado_config.reference.genome))

    if args.list:
        mikado_config = parse_list_file(mikado_config, args.list)
    elif args.gff and args.gff != [""] and args.gff != []:
        mikado_config = parse_gff_args(mikado_config, args)

    if getattr(args, "exclude_redundant", None) in (True, False):
        mikado_config.prepare.exclude_redundant = args.exclude_redundant
        mikado_config.prepare.files.exclude_redundant = [args.exclude_redundant] * len(mikado_config.prepare.files.gff)
    elif mikado_config.prepare.exclude_redundant is True:
        mikado_config.prepare.files.exclude_redundant = [True] * len(mikado_config.prepare.files.gff)
    elif not mikado_config.prepare.files.exclude_redundant:
        mikado_config.prepare.files.exclude_redundant = [False] * len(mikado_config.prepare.files.gff)
    elif len(mikado_config.prepare.files.exclude_redundant) != len(mikado_config.prepare.files.gff):
        raise InvalidConfiguration("Mismatch between exclude_redundant and gff files")

    if not mikado_config.prepare.files.reference:
        mikado_config.prepare.files.reference = [False] * len(mikado_config.prepare.files.gff)
    elif len(mikado_config.prepare.files.reference) != len(mikado_config.prepare.files.gff):
        raise InvalidConfiguration(
            "Mismatch between 'reference' and 'gff' in the prepare.files section. They must be of the same length.")

    # Set values from fields
    mikado_config.prepare.minimum_cdna_length = getattr(args, "minimum_cdna_length", None) if \
        getattr(args, "minimum_cdna_length", None) else mikado_config.prepare.minimum_cdna_length
    mikado_config.prepare.max_intron_length = getattr(args, "max_intron_length", None) if \
        getattr(args, "max_intron_length", None) else mikado_config.prepare.max_intron_length
    mikado_config.prepare.single = getattr(args, "single", None) if getattr(args, "single", None) else \
        mikado_config.prepare.single
    mikado_config.multiprocessing_method = getattr(args, "start_method", None) if \
        getattr(args, "start_method", None) else mikado_config.multiprocessing_method
    mikado_config.prepare.files.output_dir = getattr(args, "output_dir", None) if \
        getattr(args, "output_dir", None) else mikado_config.prepare.files.output_dir
    mikado_config.prepare.lenient = True if getattr(args, "lenient", None) is not None else \
        mikado_config.prepare.lenient
    mikado_config.prepare.strip_faulty_cds = True if getattr(args, "strip_faulty_cds", None) else \
        mikado_config.prepare.strip_faulty_cds
    mikado_config.prepare.strip_cds = True if getattr(args, "strip_cds", None) else mikado_config.prepare.strip_cds
    if len(mikado_config.prepare.files.strip_cds) == 0:
        mikado_config.prepare.files.strip_cds = [mikado_config.prepare.strip_cds] * len(mikado_config.prepare.files.gff)

    mikado_config.serialise.codon_table = str(args.codon_table) if (
            getattr(args, "codon_table", None) not in (None, False, True)) else mikado_config.serialise.codon_table

    if args.random_seed is True:
        mikado_config.seed = None
    elif args.seed is not None:
        mikado_config.seed = args.seed
    else:
        pass

    mikado_config.check()
    assert isinstance(mikado_config.reference.genome, str)
    return mikado_config


def setup(args, logger=None) -> (argparse.Namespace, Union[MikadoConfiguration, DaijinConfiguration],
                                 logging.Logger):
    """Method to set up the analysis using the JSON configuration
    and the command line options.

    :param args: the ArgumentParser-derived namespace.
    :param logger: the logger to use for the function call. If none is provided, a new one will be created.
    """

    if logger is None or not isinstance(logger, logging.Logger):
        logger = logging.getLogger("prepare")
        logger.setLevel(logging.INFO)

    logger.debug("Starting to get prepare arguments")
    mikado_config = load_and_validate_config(args.configuration, logger=logger)
    parse_prepare_options(args, mikado_config)

    try:
        os.makedirs(mikado_config.prepare.files.output_dir, exist_ok=True)
    except (OSError, PermissionError, FileExistsError) as exc:
        logger.error("Failed to create the output directory!")
        logger.exception(exc)
        raise exc

    if len(mikado_config.prepare.files.gff) == 0:
        parser = prepare_parser()
        logger.error("No input files found!")
        print(parser.format_help())
        sys.exit(0)

    if args.procs is not None and args.procs > 0:
        mikado_config.threads = args.procs

    mikado_config, logger = check_log_settings_and_create_logger(mikado_config, args.log, args.log_level,
                                                                 section="prepare")

    logger.info(f"Mikado version: {version.__version__}")
    logger.info(f"Command line: {' '.join(sys.argv)}")
    logger.info(f"Random seed: {mikado_config.seed}")

    mikado_config.prepare.files.out = os.path.basename(mikado_config.prepare.files.out)
    if getattr(args, "out") not in (None, False):
        mikado_config.prepare.files.out = os.path.basename(args.out)

    mikado_config.prepare.files.out_fasta = os.path.basename(args.out_fasta) if args.out_fasta is not None else \
        os.path.basename(mikado_config.prepare.files.out_fasta)

    mikado_config.reference.genome = mikado_config.reference.genome.decode() if \
        isinstance(mikado_config.reference.genome, bytes) else mikado_config.reference.genome

    return args, mikado_config, logger


def prepare_launcher(args):

    args, mikado_config, logger = setup(args)
    assert isinstance(mikado_config, (MikadoConfiguration, DaijinConfiguration))
    prepare(mikado_config, logger)
    sys.exit(0)


def prepare_parser():
    """
    This function defines the parser for the command line interface
    of the program.
    :return: an argparse.Namespace object
    :rtype: argparse.ArgumentParser
    """

    def to_cpu_count(string):
        """
        :param string: cpu requested
        :rtype: int
        """
        return max(1, int(string))

    def positive(string):
        """
        Simple function to return the absolute value of the integer of the input string.
        :param string:
        :return:
        """

        return abs(int(string))

    parser = argparse.ArgumentParser("""Script to prepare a GTF for the pipeline;
    it will perform the following operations:
    1- add the "transcript" feature
    2- sort by coordinates
    3- check the strand""")
    parser.add_argument("--fasta", "--reference", "--genome", dest="reference",
                        type=argparse.FileType(), help="Genome FASTA file. Required.")
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument("--verbose", default=None, dest="log_level", action="store_const", const="DEBUG")
    verbosity.add_argument("--quiet", default=None, dest="log_level", action="store_const", const="WARNING")
    verbosity.add_argument("-lv", "--log-level", default=None,
                         choices=["DEBUG", "INFO", "WARN", "ERROR"],
                         help="Log level. Default: derived from the configuration; if absent, INFO")
    parser.add_argument("--start-method", dest="start_method",
                        choices=["fork", "spawn", "forkserver"],
                        default=None, help="Multiprocessing start method.")
    strand = parser.add_mutually_exclusive_group()
    strand.add_argument("-s", "--strand-specific", dest="strand_specific",
                        action="store_true", default=False,
                        help="""Flag. If set, monoexonic transcripts
                        will be left on their strand rather than being
                        moved to the unknown strand.""")
    strand.add_argument("-sa", "--strand-specific-assemblies",
                        default=None,
                        type=str,
                        dest="strand_specific_assemblies",
                        help="Comma-delimited list of strand specific assemblies.")
    parser.add_argument("--list", type=argparse.FileType("r"),
                        help="""Tab-delimited file containing rows with the following format:
    <file>  <label> <strandedness(def. False)> <score(optional, def. 0)> <is_reference(optional, def. False)> <exclude_redundant(optional, def. True)> <strip_cds(optional, def. False)> <skip_split(optional, def. False)>
    "strandedness", "is_reference", "exclude_redundant", "strip_cds" and "skip_split" must be boolean values (True, False)
    "score" must be a valid floating number."""
                        )
    parser.add_argument("-l", "--log", default=None, help="Log file. Optional.")
    parser.add_argument("--lenient", action="store_true", default=None,
                        help="""Flag. If set, transcripts with only non-canonical
                        splices will be output as well.""")
    parser.add_argument("-m", "--minimum-cdna-length", default=None, dest="minimum_cdna_length", type=positive,
                        help="Minimum length for transcripts. Default: 200 bps.")
    parser.add_argument("-MI", "--max-intron-size", default=None, type=positive, dest="max_intron_length",
                        help="Maximum intron length for transcripts. Default: 1,000,000 bps.")
    parser.add_argument("-p", "--procs",
                        help="Number of processors to use (default %(default)s)",
                        type=to_cpu_count, default=None)
    parser.add_argument("-scds", "--strip_cds", action="store_true", default=False,
                        help="Boolean flag. If set, ignores any CDS/UTR segment.")
    parser.add_argument("--labels", type=str, default="",
                        help="""Labels to attach to the IDs of the transcripts of the input files,
                        separated by comma.""")
    parser.add_argument("--codon-table", dest="codon_table", default=None,
                        help="""Codon table to use. Default: 0 (ie Standard, NCBI #1, but only ATG is considered \
    a valid start codon.""")
    parser.add_argument("--single", "--single-thread", action="store_true", default=False,
                        help="Disable multi-threading. Useful for debugging.")
    parser.add_argument("-od", "--output-dir", dest="output_dir",
                        type=str, default=None,
                        help="Output directory. Default: current working directory")
    parser.add_argument("-o", "--out", default=None,
                        help="Output file. Default: mikado_prepared.gtf.")
    parser.add_argument("-of", "--out_fasta", default=None,
                        help="Output file. Default: mikado_prepared.fasta.")
    parser.add_argument("--configuration", "--json-conf", dest="configuration",
                        type=str, default="",
                        help="Configuration file.")
    parser.add_argument("-er", "--exclude-redundant", default=None,
                        dest="exclude_redundant", action="store_true",
                        help="Boolean flag. If invoked, Mikado prepare will exclude redundant models,\
ignoring the per-sample instructions.")
    cds_stripping = parser.add_mutually_exclusive_group()
    cds_stripping.add_argument("--strip-faulty-cds", default=None, action="store_true",
                        help="Flag. If set, transcripts with an incorrect CDS will be retained but \
with their CDS stripped. Default behaviour: the whole transcript will be considered invalid and discarded.")
    seed_group = parser.add_mutually_exclusive_group()
    seed_group.add_argument("--seed", type=int, default=None, help="Random seed number. Default: 0.")
    seed_group.add_argument("--random-seed", action="store_true", default=False,
                            help="Generate a new random seed number (instead of the default of 0)")
    parser.add_argument("gff", help="Input GFF/GTF file(s).", nargs="*")
    parser.set_defaults(func=prepare_launcher)
    return parser



import threading
import multiprocessing as mp


class ProcRunner(threading.Thread):

    def __init__(self, function: [mp.Process], *args, **kwargs):

        self.__function = function
        self.args = args
        self.kwargs = kwargs
        self._func = self.__function(*self.args, **self.kwargs)
        super().__init__()

    def run(self):
        self._func.run()

    def __enter__(self):
        return self

    def __exit__(self):
        self.stop()

    @property
    def func(self):
        return self._func

    def join(self, *args, **kwargs):
        if self.func._popen is not None:
            self.func.join()
            self.func.terminate()
        super().join(timeout=0.1)
        if self._tstate_lock is not None:
            assert hasattr(self._tstate_lock, "release")
            if self._tstate_lock.locked():
                self._tstate_lock.release()
        self._stop()

    def stop(self):
        self.join()
        self._stop()


#!/usr/bin/env python3
# coding: utf-8


"""
This module defines monosubloci, an intermediate step between the definition of subloci
and the final definition of loci.
A Monosublocus is characterized by its containing one and only one transcript.
"""

from .abstractlocus import Abstractlocus


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class Monosublocus(Abstractlocus):
    """Very basic class which holds a single transcript."""

    __name__ = "Monosublocus"

    # ########## Special methods ############

    def __init__(self,
                 transcript_instance=None,
                 configuration=None,
                 logger=None,
                 verified_introns=None,
                 **kwargs):

        self.counter = 0  # simple tag to avoid collisions
        Abstractlocus.__init__(self, configuration=configuration, logger=logger, verified_introns=verified_introns, **kwargs)
        # this must be defined straight away
        if transcript_instance is not None:
            Abstractlocus.add_transcript_to_locus(self, transcript_instance)
            self.tid = transcript_instance.id
            self.monoexonic = transcript_instance.monoexonic
        self.feature = "Monosublocus"
        self.parent = None
        self.attributes = {}

    # pylint: disable=arguments-differ
    def __str__(self, print_cds=True, source_in_name=True):

        raise NotImplementedError(
            """This is a container used for computational purposes only,
            it should not be printed out directly!""")

    # pylint: enable=arguments-differ

    # ########## Class instance methods ##############

    def add_transcript_to_locus(self, transcript, check_in_locus=False, **kwargs):
        """For this basic class, this method raises a NotImplementedError -
        as this container should hold only one transcript.

        :param transcript
        :param check_in_locus: flag. Ignored.
        :type check_in_locus: bool
        """

        raise NotImplementedError("In a Monosublocus there should be one and only one transcript!")

    def is_intersecting(self):
        """Not implemented: this function makes no sense for a single-transcript container."""
        raise NotImplementedError(
            "Monosubloci hold a single transcript, so intersections are not calculated.")

    # ######### Properties ############

    @property
    def id(self):
        """
        Override of the Abstractlocus method, to set the name appropriately.
        :rtype : str
        """
        if self.monoexonic is True:
            addendum = "mono"
        else:
            addendum = "multi"
        if self.counter > 0:
            addendum = "{0}.{1}".format(addendum, self.counter)
        return "{0}.{1}".format(super().id, addendum)



# coding: utf-8

"""This module contains the ORM modules necessary to create the starting DB from the input data."""

from . import blast_serializer
from . import orf
from . import junction
from . import external



import unittest
from dataclasses import asdict
import marshmallow
import tempfile
import pkg_resources
from ..configuration import configurator
from .._transcripts.scoring_configuration import SizeFilter, TargetScore
from ..configuration.configuration import *
from ..loci.abstractlocus import Abstractlocus
import pickle


class TestConfigurator(unittest.TestCase):
    def setUp(self) -> None:
        self.configuration = configurator.load_and_validate_config(None)

    def test_pickling(self):
        reloaded = pickle.loads(pickle.dumps(self.configuration))
        self.assertEqual(reloaded, self.configuration)

    def test_eval_compiled(self):
        self.configuration.scoring.requirements.expression = ["cdna_length"]
        self.configuration.scoring.requirements.parameters = dict()
        self.configuration.scoring.requirements.parameters["cdna_length"] = SizeFilter(
            name="cdna_length", value=50, operator="ge")
        self.configuration.scoring.requirements.parameters["selected_cds_length"] = SizeFilter(
            name="cdna_length", value=200, operator="ge")
        self.configuration.scoring.requirements._check_my_requirements()
        values = {"cdna_length": 100, "selected_cds_length": 150}
        evaluated = dict()
        evaluated["cdna_length"] = Abstractlocus.evaluate(
            values["cdna_length"],
            self.configuration.scoring.requirements.parameters["cdna_length"])
        self.assertTrue(evaluated["cdna_length"])
        evaluated["selected_cds_length"] = Abstractlocus.evaluate(
            values["selected_cds_length"],
            self.configuration.scoring.requirements.parameters["selected_cds_length"])
        self.configuration.scoring.requirements._check_my_requirements()
        self.assertFalse(evaluated["selected_cds_length"])
        self.assertTrue(eval(self.configuration.scoring.requirements.compiled))
        # Now we are changing the contents of the expression
        self.configuration.scoring.requirements.expression = ["selected_cds_length"]
        self.configuration.scoring.requirements._check_my_requirements()
        self.assertFalse(eval(self.configuration.scoring.requirements.compiled))

    def test_target_score(self):
        """The order of the Union matters for loading!"""

        for value in (0.3, True, False, 10, 100.4):
            with self.subTest(value=value, msg=f"Testing value {value}"):
                schema = {"rescaling": "target", "value": value}
                loaded = TargetScore.Schema().load(schema)
                self.assertEqual(loaded.value, schema["value"])
                if isinstance(value, bool):
                    self.assertIsInstance(loaded.value, bool)
                else:
                    self.assertIsInstance(loaded.value, float)

        for value in ("hello", dict(), []):
            with self.subTest(value=value, msg=f"Testing value {value}"):
                schema = {"rescaling": "target", "value": value}
                with self.assertRaises(marshmallow.ValidationError):
                    TargetScore.Schema().load(schema)

    def test_codon_loading(self):
        for value in [0, "0", "1", 1, "Standard", "Flatworm Mitochondrial"]:
            with self.subTest(value=value, msg=f"Testing value {value}"):
                schema = {"codon_table": value, "max_regression": 0.2, "substitution_matrix": "blosum62"}
                loaded = SerialiseConfiguration.Schema().load(schema)
                if value in [0, "0", "1", 1]:
                    self.assertIsInstance(loaded.codon_table, int)
                else:
                    self.assertIsInstance(loaded.codon_table, str)

        for value in ["8", 8, "Universal"]:
            with self.subTest(value=value, msg=f"Testing value {value}"):
                schema = {"codon_table": value, "max_regression": 0.2, "substitution_matrix": "blosum62"}
                with self.assertRaises(marshmallow.ValidationError):
                    SerialiseConfiguration.Schema().load(schema)

    def test_load_plant_scoring(self):
        mammalian = MikadoConfiguration(pick=PickConfiguration(scoring_file="mammalian.yaml"))
        plant = MikadoConfiguration(pick=PickConfiguration(scoring_file="plant.yaml"))
        insect = MikadoConfiguration(pick=PickConfiguration(scoring_file="HISTORIC/insects.yaml"))

        self.assertTrue(mammalian.scoring.requirements.expression == plant.scoring.requirements.expression
                        != insect.scoring.requirements.expression)
        self.assertTrue(mammalian.scoring.requirements.compiled == plant.scoring.requirements.compiled
                        != insect.scoring.requirements.compiled)
        self.assertTrue(mammalian.scoring.requirements.parameters != plant.scoring.requirements.parameters)

    def test_load_invalid_scoring(self):
        erroneous = tempfile.NamedTemporaryFile(suffix=".yaml", mode="wt")
        plant = MikadoConfiguration(pick=PickConfiguration(scoring_file="plant.yaml"))
        err_scoring = copy.deepcopy(plant.scoring)
        key = next(iter(err_scoring.scoring.keys()))
        err_scoring.scoring[key].rescaling = "invalid"
        key = next(iter(err_scoring.requirements.parameters.keys()))
        del err_scoring.requirements.parameters[key]
        yaml.dump(asdict(err_scoring), erroneous)
        erroneous.flush()
        plant.pick.scoring_file = erroneous.name
        logger = create_default_logger("test_load_invalid_scoring", level="DEBUG")
        with self.assertRaises(InvalidConfiguration):
            plant.load_scoring(logger=logger)
        current = os.getcwd()
        os.chdir(os.path.dirname(erroneous.name))
        plant.scoring_file = os.path.basename(erroneous.name)
        self.assertFalse(os.path.exists("plant.yaml"))
        os.link(erroneous.name, "plant.yaml")
        plant._loaded_scoring = None
        plant.pick.scoring_file = "plant.yaml"
        with self.assertRaises(InvalidConfiguration):
            plant.load_scoring(logger=logger)
        os.remove(os.path.join(tempfile.gettempdir(), "plant.yaml"))
        plant._loaded_scoring = None
        plant.load_scoring(logger=logger)
        self.assertEqual(plant._loaded_scoring,
                         pkg_resources.resource_filename("Mikado.configuration", os.path.join("scoring_files",
                                                                                              "plant.yaml")))
        os.chdir(current)



import re
import unittest
from .. import parsers
from ..transcripts import Transcript
from ..utilities import log_utils


__author__ = 'Luca Venturini'


class MultOrfTester(unittest.TestCase):

    logger = log_utils.create_default_logger("mult_orf")

    tr_gff = """
    Chr1    TAIR10    mRNA    5928    8737    .    -    .    ID=AT1G01020.1;Parent=AT1G01020;Name=AT1G01020.1;Index=1
    Chr1    TAIR10    five_prime_UTR    8667    8737    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    8571    8666    .    -    0    Parent=AT1G01020.1;
    Chr1    TAIR10    exon    8571    8737    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    8417    8464    .    -    0    Parent=AT1G01020.1;
    Chr1    TAIR10    exon    8417    8464    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    8236    8325    .    -    0    Parent=AT1G01020.1;
    Chr1    TAIR10    exon    8236    8325    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    7942    7987    .    -    0    Parent=AT1G01020.1;
    Chr1    TAIR10    exon    7942    7987    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    exon    7762    7835    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    7564    7649    .    -    0    Parent=AT1G01020.1;
    Chr1    TAIR10    exon    7564    7649    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    7384    7450    .    -    1    Parent=AT1G01020.1;
    Chr1    TAIR10    exon    7384    7450    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    7157    7232    .    -    0    Parent=AT1G01020.1;
    Chr1    TAIR10    exon    7157    7232    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    CDS    6915    7069    .    -    2    Parent=AT1G01020.1;
    Chr1    TAIR10    three_prime_UTR    6437    6914    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    exon    6437    7069    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    three_prime_UTR    5928    6263    .    -    .    Parent=AT1G01020.1
    Chr1    TAIR10    exon    5928    6263    .    -    .    Parent=AT1G01020.1"""

    tr_lines = [line.strip() for line in tr_gff.split("\n") if line.strip() != '']
    for pos, line in enumerate(tr_lines):
        tr_lines[pos] = re.sub(r"\s+", r"\t", line)
        assert len(tr_lines[pos].split("\t")) == 9, line.split("\t")

    tr_gff_lines = [parsers.GFF.GffLine(line) for line in tr_lines]

    for l in tr_gff_lines:
        assert l.header is False

    def setUp(self):
        """Basic creation test."""

        self.tr = Transcript(self.tr_gff_lines[0], logger=self.logger)
        for line in self.tr_gff_lines[1:]:
            self.tr.add_exon(line)

    def test_missing_cds(self):

        with self.assertLogs("mult_orf", level="WARNING") as cm:
            self.tr.finalize()
        self.assertFalse(self.tr.is_coding)

    def test_split_cds_exon_one(self):

        lines = """
        Chr1    TAIR10    CDS    7762    7800    .    -    .    Parent=AT1G01020.1
        Chr1    TAIR10    CDS    7810    7835    .    -    .    Parent=AT1G01020.1
        """

        lines = ["\t".join(line.strip().split()) for line in lines.split("\n") if line.strip() != '']
        for line in lines:
            line = parsers.GFF.GffLine(line)
            self.assertFalse(line.header)
            self.assertEqual(line.parent[0], self.tr.id)
            self.tr.add_exon(line)
        with self.assertLogs("mult_orf", level="WARNING") as cm:
            self.tr.finalize()
        self.assertFalse(self.tr.is_coding)

    def test_split_cds_exon_two(self):

        lines = """
        Chr1    TAIR10    CDS    7762    7800    .    -    .    Parent=AT1G01020.1
        Chr1    TAIR10    CDS    7800    7835    .    -    .    Parent=AT1G01020.1
        """

        lines = ["\t".join(line.strip().split()) for line in lines.split("\n") if line.strip() != '']
        for line in lines:
            line = parsers.GFF.GffLine(line)
            self.assertFalse(line.header)
            self.tr.add_exon(line)
        with self.assertLogs("mult_orf", level="WARNING") as cm:
            self.tr.finalize()
        self.assertFalse(self.tr.is_coding)

    def test_restore(self):
        lines = """
        Chr1    TAIR10    CDS    7762    7835    .    -    0    Parent=AT1G01020.1
        """

        lines = ["\t".join(line.strip().split()) for line in lines.split("\n") if line.strip() != '']
        for line in lines:
            line = parsers.GFF.GffLine(line)
            self.tr.add_exon(line)
        self.assertTrue(self.tr.is_coding)


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple
from .transcript import Transcript
from ..exceptions import InvalidCDS


class TranscriptComputer(Transcript):
    """
    Class that is used to calculate and store basic statistics about a transcript object.
    """

    data_fields = ["parent", 'chrom',
                   'start', 'end',
                   'introns', 'exons',
                   'exon_lengths', 'intron_lengths',
                   'cdna_length', 'selected_cds_length',
                   'cds_intron_lengths', 'cds_exon_lengths',
                   "five_utr_length", "three_utr_length",
                   "five_utr_num", "three_utr_num",
                   "selected_end_distance_from_junction"]
    data_tuple = namedtuple("transcript_data", data_fields)

    def __init__(self, *args, **kwargs):
        kwargs.pop("accept_undefined_multi", None)
        kwargs.pop("trust_orf", None)
        super().__init__(accept_undefined_multi=True, trust_orf=True, *args, **kwargs)
        self.exon_lengths = []
        self.cds_exon_lengths = []
        self.utr_exon_lengths = []

        self.intron_lengths = []
        self.cds_intron_lengths = []
        self.utr_intron_lengths = []

    def finalize(self):
        """
        Method to be called when all exons/features have been
        added to the transcript. It will call the parent's finalize method,
        followed by calculation of the necessary statistics.
        """
        try:
            super().finalize()
        except InvalidCDS:
            super().strip_cds()

        self.exon_lengths = [e[1] - e[0] + 1 for e in self.exons]
        self.cds_exon_lengths = [c[1] - c[0] + 1 for c in self.selected_cds]
        self.utr_exon_lengths = [u[1] - u[0] + 1 for u in self.three_utr + self.five_utr]

        self.intron_lengths = [i[1] - i[0] + 1 for i in self.introns]
        self.cds_intron_lengths = [i[1] - i[0] for i in self.selected_cds_introns]
        self.utr_intron_lengths = [i[1] - i[0] for i in self.introns if
                                   i not in self.selected_cds_introns]
